from .core import *
from .calculate_od import *
from .conversion_utils import *
from .network_clean import *
from .optimization import *
from .osm_parser import *

# make sure gdal, geopy, and boltons are installed first
from .load_osm import *



#the cleaning network part
import os, sys, time

#import GOSTnet as gn
import importlib
#import networkx as nx
import osmnx as ox
from shapely.ops import unary_union
from shapely.wkt import loads
from shapely.geometry import LineString, MultiLineString, Point


def CleanNetwork(G, wpath = '', country='', UTM={'init': 'epsg:3857'}, WGS = {'init': 'epsg:4326'}, junctdist = 50, verbose = False):   
    ''' Topologically simplifies an input graph object by collapsing junctions and removing interstital nodes
    # REQUIRED - G: a graph object containing nodes and edges. edges should have a property 
    #               called 'Wkt' containing geometry objects describing the roads
    #            wpath: the write path - a drive directory for inputs and output
    #            country: this parameter allows for the sequential processing of multiple countries
    #            UTM: the epsg code of the projection, in metres, to apply the junctdist
    # OPTIONAL - junctdist: distance within which to collapse neighboring nodes. simplifies junctions. 
    #            Set to 0.1 if not simplification desired. 50m good for national (primary / secondary) networks
    #            verbose: if True, saves down intermediate stages for dissection
    '''
    # Squeezes clusters of nodes down to a single node if they are within the snapping tolerance
    a = gn.simplify_junctions(G, UTM, WGS, junctdist)

    # ensures all streets are two-way
    a = gn.add_missing_reflected_edges(a)
    
    #save progress
    if verbose is True: 
        gn.save(a, 'a', wpath)
    
    # Finds and deletes interstital nodes based on node degree
    b = gn.custom_simplify(a)
    
    # rectify geometry
    for u, v, data in b.edges(data = True):
        if type(data['Wkt']) == list:
                data['Wkt'] = gn.unbundle_geometry(data['Wkt'])
    
    # save progress
    if verbose is True: 
        gn.save(b, 'b', wpath)
    
    # For some reason CustomSimplify doesn't return a MultiDiGraph. Fix that here
    c = gn.convert_to_MultiDiGraph(b)

    # This is the most controversial function - removes duplicated edges. This takes care of two-lane but separate highways, BUT
    # destroys internal loops within roads. Can be run with or without this line
    c = gn.remove_duplicate_edges(c)

    # Run this again after removing duplicated edges
    c = gn.custom_simplify(c)

    # Ensure all remaining edges are duplicated (two-way streets)
    c = gn.add_missing_reflected_edges(c)
    
    # save final
    if verbose:
        gn.save(c, '%s_processed' % country, wpath)
    
    print('Edge reduction: %s to %s (%d percent)' % (G.number_of_edges(), 
                                               c.number_of_edges(), 
                                               ((G.number_of_edges() - c.number_of_edges())/G.number_of_edges()*100)))
    return c



####################################################################################################
# Load OSM data into network graph
# Benjamin Stewart and Charles Fox
# Purpose: take an input dataset as a OSM file and return a network object
####################################################################################################

import os, sys, time

import shapely.ops

import geopandas as gpd
import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from osgeo import ogr
from rtree import index
from shapely import speedups
from shapely.geometry import LineString, MultiLineString, MultiPoint, Point
from geopy.distance import vincenty
from boltons.iterutils import pairwise
from shapely.wkt import loads,dumps

class OSM_to_network(object):
    """
    Object to load OSM PBF to networkX objects. \
    EXAMPLE: \
    G_loader = losm.OSM_to_network(bufferedOSM_pbf) \
    G_loader.generateRoadsGDF() \
    G = G.initialReadIn() \

    snap origins and destinations \
    o_snapped = gn.pandana_snap(G, origins) \
    d_snapped = gn.pandana_snap(G, destinations) \
    """

    def __init__(self, osmFile):
        ''' Generate a networkX object from a osm file
        '''
        self.osmFile = osmFile
        self.roads_raw = self.fetch_roads(osmFile)

    def generateRoadsGDF(self, in_df = None, outFile='', verbose = False):
        ''' Convert the raw OSM roads from the OSM file to a GeoDataFrame
        '''
        if type(in_df) != gpd.geodataframe.GeoDataFrame:
            in_df = self.roads_raw
        roads = self.get_all_intersections(in_df, verboseness = verbose)
        roads['key'] = ['edge_'+str(x+1) for x in range(len(roads))]
        np.arange(1,len(roads)+1,1)

        def get_nodes(x):
            return list(x.geometry.coords)[0],list(x.geometry.coords)[-1]

        nodes = gpd.GeoDataFrame(roads.apply(lambda x: get_nodes(x),axis=1).apply(pd.Series))
        nodes.columns = ['u','v']

        roads['length'] = roads.geometry.apply(lambda x : self.line_length(x))
        roads.rename(columns={'geometry':'Wkt'}, inplace=True)
        roads = pd.concat([roads,nodes],axis=1)

        if outFile != '':
            roads.to_csv(outFile)

        self.roadsGPD = roads

    def filterRoads(self, acceptedRoads = ['primary','primary_link','secondary','secondary_link','motorway','motorway_link','trunk','trunk_link']):
        ''' Extract certain times of roads from the OSM before the netowrkX conversion 
        
        INPUT
            [ optional ] acceptedRoads [ list of strings ] 
        
        RETURNS
            None - the raw roads are filtered based on the list of accepted roads
        '''
        self.roads_raw = self.roads_raw.loc[self.roads_raw.infra_type.isin(acceptedRoads)]

    def fetch_roads(self, data_path):
        ''' Extract roads from osm pbf
        
        RETURNS
           None
        '''
        if data_path.split('.')[-1] == 'pbf':
            driver = ogr.GetDriverByName("OSM")
            data = driver.Open(data_path)
            sql_lyr = data.ExecuteSQL("SELECT osm_id,highway FROM lines WHERE highway IS NOT NULL")
            roads=[]

            for feature in sql_lyr:
                if feature.GetField("highway") is not None:
                    osm_id = feature.GetField("osm_id")
                    shapely_geo = loads(feature.geometry().ExportToWkt())
                    if shapely_geo is None:
                        continue
                    highway=feature.GetField("highway")
                    roads.append([osm_id,highway,shapely_geo])

            if len(roads) > 0:
                road_gdf = gpd.GeoDataFrame(roads,columns=['osm_id','infra_type','geometry'],crs={'init': 'epsg:4326'})
                return road_gdf

        elif data_path.split('.')[-1] == 'shp':
            road_gdf = gpd.read_file(data_path)
            return road_gdf

        else:
            print('No roads found')

    def line_length(self, line, ellipsoid='WGS-84'):
        '''Length of a line in meters, given in geographic coordinates

        Adapted from https://gis.stackexchange.com/questions/4022/looking-for-a-pythonic-way-to-calculate-the-length-of-a-wkt-linestring#answer-115285

        Arguments:
            line {Shapely LineString} -- a shapely LineString object with WGS-84 coordinates
            ellipsoid {String} -- string name of an ellipsoid that `geopy` understands (see
                http://geopy.readthedocs.io/en/latest/#module-geopy.distance)

        Returns:
            Length of line in meters
        '''
        if line.geometryType() == 'MultiLineString':
            return sum(line_length(segment) for segment in line)

        return sum(
                    vincenty(tuple(reversed(a)), tuple(reversed(b)), ellipsoid=ellipsoid).kilometers
                    for a, b in pairwise(line.coords)
        )

    def get_all_intersections(self, shape_input, idx_osm = None, verboseness = False):
        # Initialize Rtree
        idx_inters = index.Index()
        # Load data
        #all_data = dict(zip(list(shape_input.osm_id),list(shape_input.geometry),list(shape_input.infra_type)))
        ### TODO - it shouldn't be necessary to reference the geometry column specifically
        #   ... but here we are
        if idx_osm is None:
            idx_osm = shape_input['geometry'].sindex

        # Find all the intersecting lines to prepare for cutting
        count = 0
        tLength = shape_input.shape[0]
        inters_done = {}
        new_lines = []
        allCounts = []

        for idx, row in shape_input.iterrows():
            key1 = row.osm_id
            line = row.geometry
            infra_type = row.infra_type
            if count % 1000 == 0 and verboseness == True:
                print("Processing %s of %s" % (count, tLength))
            count += 1
            intersections = shape_input.iloc[list(idx_osm.intersection(line.bounds))]
            intersections = dict(zip(list(intersections.osm_id),list(intersections.geometry)))
            if key1 in intersections: 
                intersections.pop(key1)
            # Find intersecting lines
            for key2,line2 in intersections.items():
                # Check that this intersection has not been recorded already
                if (key1, key2) in inters_done or (key2, key1) in inters_done:
                    continue
                # Record that this intersection was saved
                inters_done[(key1, key2)] = True
                # Get intersection
                if line.intersects(line2):
                    # Get intersection
                    inter = line.intersection(line2)
                    # Save intersecting point
                    if "Point" == inter.type:
                        idx_inters.insert(0, inter.bounds, inter)
                    elif "MultiPoint" == inter.type:
                        for pt in inter:
                            idx_inters.insert(0, pt.bounds, pt)

            # cut lines where necessary and save all new linestrings to a list
            hits = [n.object for n in idx_inters.intersection(line.bounds, objects=True)]

            if len(hits) != 0:
                try:
                    out = shapely.ops.split(line, MultiPoint(hits))
                    new_lines.append([{'geometry': LineString(x), 'osm_id':key1,'infra_type':infra_type} for x in out.geoms])
                except:
                    pass
            else:
                new_lines.append([{'geometry': line, 'osm_id':key1,
                        'infra_type':infra_type}])

        # Create one big list and treat all the cutted lines as unique lines
        flat_list = []
        all_data = {}

        #item for sublist in new_lines for item in sublist
        i = 1
        for sublist in new_lines:
            if sublist is not None:
                for item in sublist:
                    item['id'] = i
                    flat_list.append(item)
                    i += 1
                    all_data[i] = item

        # Transform into geodataframe and add coordinate system
        full_gpd = gpd.GeoDataFrame(flat_list,geometry ='geometry')
        full_gpd.crs = {'init' :'epsg:4326'}
        return(full_gpd)

    def initialReadIn(self, fpath=None, wktField='Wkt'):
        ''' Convert the OSM object to a networkX object
        
        INPUT
        [ optional ] fpath [ string ] - path to CSV file with roads to read in
        
        RETURNS
            [ Networkx Multi-digraph ]
        '''
        if isinstance(fpath, str):
            edges_1 = pd.read_csv(fpath)
            edges_1 = edges_1[wktField].apply(lambda x: loads(x))
        elif isinstance(fpath, gpd.GeoDataFrame):
            edges_1 = fpath
        else:
            try:
                edges_1 = self.roadsGPD
            except:
                self.generateRoadsGDF()
                edges_1 = self.roadsGPD
        edges = edges_1.copy()
        node_bunch = list(set(list(edges['u']) + list(edges['v'])))
        def convert(x):
            u = x.u
            v = x.v
            data = {'Wkt':x.Wkt,
                   'id':x.id,
                   'infra_type':x.infra_type,
                   'osm_id':x.osm_id,
                   'key': x.key,
                   'length':x.length}
            return (u, v, data)

        edge_bunch = edges.apply(lambda x: convert(x), axis = 1).tolist()
        G = nx.MultiDiGraph()
        G.add_nodes_from(node_bunch)
        G.add_edges_from(edge_bunch)
        for u, data in G.nodes(data = True):
            if type(u) == str:
                q = tuple(float(x) for x in u[1:-1].split(','))
            if type(u) == tuple:
                q = u
            data['x'] = q[0]
            data['y'] = q[1]
        G = nx.convert_node_labels_to_integers(G)
        self.network = G
        return G



import time
import multiprocessing as mp
from pulp import LpInteger, LpVariable, LpProblem, lpSum, LpMaximize
import pandas

speed_dict = {
                'residential': 20,  # kmph
                'primary': 40, # kmph
                'primary_link':35,
                'motorway':45,
                'motorway_link': 40,
                'trunk': 40,
                'trunk_link':35,
                'secondary': 30, # kmph
                'secondary_link':25,
                'tertiary':30,
                'tertiary_link': 25,
                'unclassified':20,
                'road':20,
                'crossing':20,
                'living_street':20
                }


def optimize_facility_locations(OD, facilities, p, existing_facilities = None, verbose=False, execute=True, write=''):
    """
    Function for identifying spatially optimal locations of facilities (P-median problem)

    :param OD: an Origin:Destination matrix, origins as rows, destinations as columns, in pandas DataFrame format. 
    :param facilities: The 'destinations' of the OD-Matrix. MUST be a list of objects included in OD.columns (or subset) if certain nodes are unsuitable for facility locations
    :param p: the number of facilities to solve for
    :param existing_facilities: facilities to always include in the solution. MUST be in 'facilities' list
    :param verbose: print a bunch of status updates
    :param execute: should the problem be executed
    :param write: outPath to write problem
    """
    
    num_procs = mp.cpu_count()
    if verbose:
        print('cpu count: %s' % num_procs)
    t1=time.time()
    
    if type(OD) != pandas.core.frame.DataFrame:
        raise ValueError('OD must be pandas Dataframe!')

    for f in facilities:
        if f not in OD.columns:
            raise ValueError('Potential facility locations MUST be in OD.columns')

    if p < 1:
        raise ValueError('need to solve for more than one facility!')
    elif p > len(facilities):
        raise ValueError('need to solve for fewer locations than location options!')
    
    if verbose:
        print('Setting up problem')
    
    origins = OD.index
    origins = list(map(int, origins))

    X = pulp.LpVariable.dicts('X',(facilities),0,1,pulp.LpInteger)

    Y = pulp.LpVariable.dicts('Y', (origins,facilities),0,1,pulp.LpInteger)

    prob = pulp.LpProblem('P Median', pulp.LpMinimize)

    prob += sum(sum(OD.loc[i,j] * Y[i][j] for j in facilities) for i in origins)

    prob += pulp.lpSum([X[j] for j in facilities]) == p

    for i in origins: prob += sum(Y[i][j] for j in facilities) == 1

    for i in origins:
        for j in facilities:
            prob +=  Y[i][j] <= X[j]

    if existing_facilities is not None:
        for e in existing_facilities:
            prob += X[e] == 1
    
    if verbose:
        print('Set up the problem')
    
    if write != '':
        prob.writeLP(write)
        
    if execute:
        prob.solve()
        #the code below needs coin-or to be installed in the OS
        #prob.solve(pulp.COIN_CMD(threads=16)))
        ans = []
        if verbose:
            print("Processing time took: ", time.time() - t1)
        
        for v in prob.variables():
            subV = v.name.split('_')

            if subV[0] == "X" and v.varValue == 1:
                ans.append(int(str(v).split('_')[1]))

        return ans
    else:
        return(prob)



def optimize_set_coverage(OD, max_coverage = 2000, existing_facilities = None):

    ### Determine the minimum number of facilities and their locations in order to cover all demands within a pre-specified maximum distance (or time) coverage (Location Set-Covering Problem) ###
    # REQUIRED:   OD - an Origin:Destination matrix, origins as rows, destinations
    #             as columns, in pandas DataFrame format.
    #             max_coverage - The pre-specified maximum distance (or time) coverage.
    # OPTIONAL:   existing_facilities - facilities to always include in the
    #             solution. MUST be in 'facilities' list
    # -------------------------------------------------------------------------#

    # OD keys must be integers
    OD.columns = OD.columns.astype(int)

    origins = OD.index
    origins = list(map(int, origins))

    facilities = OD.keys()
    facilities = list(map(int, facilities))

    X = LpVariable.dicts('X',(facilities),0,1,LpInteger)

    prob = LpProblem('Set Cover', LpMinimize)

    prob += sum(X[j] for j in facilities)

    for i in origins:
        #set of facilities that are eligible to provide coverage to point i
        eligibleFacilities = []
        for j in facilities:
            if OD.loc[i,j] <= max_coverage:
                eligibleFacilities.append(j)
        prob += sum(X[j] for j in eligibleFacilities) >= 1

    prob.solve()

    ans = []

    for v in prob.variables():
        subV = v.name.split('_')

        if subV[0] == "X" and v.varValue == 1:
            ans.append(int(str(v).split('_')[1]))

    if existing_facilities is not None:
        for e in existing_facilities:
            prob += X[e] == 1

    #print out other variables
    print('number of origins')
    print(len(origins))

    totalCoveredFacilities = 0

    for i in origins:
        coveredFacilities = []
        for j in ans:
            if OD.loc[i,j] <= max_coverage:
                coveredFacilities.append(j)
        if len(coveredFacilities) >= 1:
            totalCoveredFacilities += 1

    print('print totalCoveredFacilities')
    print(totalCoveredFacilities)
            
    print('print percent coverage')
    print(totalCoveredFacilities/len(origins)*100)
    
    print('print prob obj')
    print(prob.objective)

    return ans

def optimize_partial_set_coverage(OD, pop_coverage = .8, max_coverage = 2000, origins_pop_series = None, existing_facilities = None):

    ### Function to determine the minimum number of facilities and their locations in order to cover a given fraction of the population within a pre-specified maximum distance (or time) coverage (Partial Set-Covering Problem). Do not use a demand-weighted OD matrix as an input. ###
    # REQUIRED:   OD - an Origin:Destination matrix, origins as rows, destinations
    #             as columns, in pandas DataFrame format.
    #             max_coverage - The pre-specified maximum distance (or time) coverage.
    #             pop_coverage - The given fraction of the population that should be covered
    # OPTIONAL:   origins_pop_series - a series that contains each origin as the key, and each origin's population as the value
    #             existing_facilities - facilities to always include in the
    #             solution. MUST be in 'facilities' list
    # -------------------------------------------------------------------------#

    # OD keys must be integers
    OD.columns = OD.columns.astype(int)
    
    origins = OD.index
    origins = list(map(int, origins))

    facilities = OD.keys()
    facilities = list(map(int, facilities))  

    X = LpVariable.dicts('X',(facilities),0,1,LpInteger)

    Z = LpVariable.dicts('Z',(origins),0,1,LpInteger)

    prob = LpProblem('Partial Set Cover', LpMinimize)

    #objective function
    prob += sum(X[j] for j in facilities)

    for i in origins:
        #set of facilities that are eligible to provide coverage to point i
        eligibleFacilities = []
        for j in facilities:
            if OD.loc[i,j] <= max_coverage:
                eligibleFacilities.append(j)
        # corrected formulation
        prob += sum(X[j] for j in eligibleFacilities) - Z[i] >= 0


    #if origins_pop_series exists then sum up total population and multiply by pop_coverage

    if origins_pop_series is not None:

        #print('print origins_pop_series')
        #print(origins_pop_series)

        #print('print sum(origins_pop_series)')
        #print(sum(origins_pop_series))

        #print('print origins')
        #print(origins)

        #for i in origins:
            #print(i)

        min_coverage = sum(origins_pop_series) * pop_coverage

        print('print min_coverage')
        print(min_coverage)

        prob += sum(Z[i] * origins_pop_series[i] for i in origins) >= min_coverage

    else:

        min_coverage = len(origins) * pop_coverage

        print('print min_coverage')
        print(min_coverage)

        prob += sum(Z[i] for i in origins) >= min_coverage

    if existing_facilities is not None:
        for e in existing_facilities:
            prob += X[e] == 1

    prob.solve()

    ans = []

    for v in prob.variables():
        subV = v.name.split('_')

        if subV[0] == "X" and v.varValue == 1:
            ans.append(int(str(v).split('_')[1]))

    return ans

def optimize_max_coverage(OD, p_facilities = 5, max_coverage = 2000, origins_pop_series = None, existing_facilities = None):

    ### Determine the location of P facilities in order to maximize the demand covered within a pre-specified maximum distance coverage (Max Cover). Do not use a demand-weighted OD matrix as an input. ###
    # REQUIRED:   OD - an Origin:Destination matrix, origins as rows, destinations
    #             as columns, in pandas DataFrame format.
    #             max_coverage - The pre-specified maximum distance (or time) coverage.
    #             p_facilities - The number of facilities to locate.
    # OPTIONAL:   existing_facilities - facilities to always include in the
    #             solution. MUST be in 'facilities' list
    # -------------------------------------------------------------------------#

    # OD keys must be integers
    OD.columns = OD.columns.astype(int)
    
    origins = OD.index
    origins = list(map(int, origins))

    facilities = OD.keys()
    facilities = list(map(int, facilities))  

    #If a facility is located at candidate site j
    X = LpVariable.dicts('X', (facilities),0,1,LpInteger)

    #If demand Y is covered
    Y = LpVariable.dicts('Y', (origins),0,1,LpInteger)

    prob = LpProblem('Max Cover', LpMaximize)

    if origins_pop_series is not None:

        #objective function
        prob += sum(origins_pop_series[i] * Y[i] for i in origins)

    else:

        #objective function
        prob += sum(Y[i] for i in origins)

    for i in origins:
        #set of facilities that are eligible to provide coverage to point i
        eligibleFacilities = []
        for j in facilities:
            if OD.loc[i,j] <= max_coverage:
                eligibleFacilities.append(j)
        prob += sum(X[j] for j in eligibleFacilities) >= Y[i]

    prob += sum(X[j] for j in facilities) == p_facilities

    if existing_facilities is not None:
        for e in existing_facilities:
            prob += X[e] == 1

    prob.solve()

    #print('print prob')
    #print(prob)

    ans = []

    for v in prob.variables():
        subV = v.name.split('_')

        if subV[0] == "X" and v.varValue == 1:
            ans.append(int(str(v).split('_')[1]))

    print('print objective value')
    print(prob.objective.value())
    #print(prob.objective)

    return ans



import os, sys, logging, warnings, time

import pyproj
import networkx as nx
import osmnx as ox
import pandas as pd
import geopandas as gpd
import numpy as np

from scipy import spatial
from functools import partial
from shapely.wkt import loads
from shapely.geometry import Point, LineString, MultiLineString, box
from shapely.ops import linemerge, unary_union, transform
from collections import Counter

def combo_csv_to_graph(fpath, u_tag = 'u', v_tag = 'v', geometry_tag = 'Wkt', largest_G = False):
    """
    Function for generating a G object from a saved combo .csv

    :param fpath: path to a .csv containing edges (WARNING: COMBO CSV only)
    :param u_tag: specify column containing u node ID if not labelled 'u'
    :param v_tag: specify column containing u node ID if not labelled 'v'
    :param geometry_tag: specify column containing u node ID if not
    :returns: a multidigraph object
    """

    edges_1 = pd.read_csv(os.path.join(fpath))

    edges = edges_1.copy()

    node_bunch = list(set(list(edges[u_tag]) + list(edges[v_tag])))

    col_list = list(edges.columns)
    drop_cols = [u_tag, v_tag, geometry_tag]
    attr_list = [col_entry for col_entry in col_list if col_entry not in drop_cols]

    def convert(x, attr_list):
        u = x[u_tag]
        v = x[v_tag]
        data = {'Wkt':loads(x[geometry_tag])}
        for i in attr_list:
            data[i] = x[i]

        return (u, v, data)

    edge_bunch = edges.apply(lambda x: convert(x, attr_list), axis = 1).tolist()

    G = nx.MultiDiGraph()

    G.add_nodes_from(node_bunch)
    G.add_edges_from(edge_bunch)

    for u, data in G.nodes(data = True):
        q = tuple(float(x) for x in u[1:-1].split(','))
        data['x'] = q[0]
        data['y'] = q[1]

    G = nx.convert_node_labels_to_integers(G)

    if largest_G == True:
        list_of_subgraphs = list(nx.strongly_connected_component_subgraphs(G))
        l = 0
        cur_max = 0
        for i in list_of_subgraphs:
            if i.number_of_edges() > cur_max:
                cur_max = i.number_of_edges()
                max_ID = l
            l +=1
        G = list_of_subgraphs[max_ID]

    return G

def edges_and_nodes_csv_to_graph(fpath_nodes, fpath_edges, u_tag = 'stnode', v_tag = 'endnode', geometry_tag = 'Wkt', largest_G = False):

    """
    Function for generating a G object from a saved .csv of edges

    :param fpath_nodes: 
      path to a .csv containing nodes
    :param fpath_edges: 
      path to a .csv containing edges
    :param u_tag: 
      optional. specify column containing u node ID if not labelled 'stnode'
    :param v_tag: 
      specify column containing v node ID if not labelled 'endnode'
    :param geometry_tag: 
      specify column containing geometry if not labelled 'Wkt'
    :returns: 
      a multidigraph object
    """

    nodes_df = pd.read_csv(fpath_nodes)
    edges_df = pd.read_csv(fpath_edges)

    chck_set = list(set(list(edges_df[u_tag]) + list(edges_df[v_tag])))

    def check(x, chck_set):
        if x in chck_set:
            return 1
        else:
            return 0

    nodes_df['chck'] = nodes_df['node_ID'].apply(lambda x: check(x, chck_set))

    nodes_df = nodes_df.loc[nodes_df.chck == 1]

    def convert_nodes(x):
        u = x.node_ID
        data = {'x':x.x,
               'y':x.y}
        return (u, data)

    node_bunch = nodes_df.apply(lambda x: convert_nodes(x), axis = 1).tolist()

    col_list = list(edges_df.columns)
    drop_cols = [u_tag, v_tag, geometry_tag]
    attr_list = [col_entry for col_entry in col_list if col_entry not in drop_cols]

    def convert_edges(x):
        u = x[u_tag]
        v = x[v_tag]
        data = {'Wkt':loads(x[geometry_tag])}
        for i in attr_list:
            data[i] = x[i]

        return (u, v, data)

    edge_bunch = edges_df.apply(lambda x: convert_edges(x), axis = 1).tolist()

    G = nx.MultiDiGraph()

    G.add_nodes_from(node_bunch)
    G.add_edges_from(edge_bunch)

    G = nx.convert_node_labels_to_integers(G)

    if largest_G == True:
        list_of_subgraphs = list(nx.strongly_connected_component_subgraphs(G))
        l = 0
        cur_max = 0
        for i in list_of_subgraphs:
            if i.number_of_edges() > cur_max:
                cur_max = i.number_of_edges()
                max_ID = l
            l +=1
        G = list_of_subgraphs[max_ID]

    return G

def node_gdf_from_graph(G, crs = {'init' :'epsg:4326'}, attr_list = None, geometry_tag = 'geometry', xCol='x', yCol='y'):
    """
    Function for generating GeoDataFrame from Graph

    :param G: a graph object G
    :param crs: projection of format {'init' :'epsg:4326'}. Defaults to WGS84. note: here we are defining the crs of the input geometry - we do NOT reproject to this crs. To reproject, consider using geopandas' to_crs method on the returned gdf.
    :param attr_list: list of the keys which you want to be moved over to the GeoDataFrame, if not all. Defaults to None, which will move all.
    :param geometry_tag: specify geometry attribute of graph, default 'geometry'
    :param xCol: if no shapely geometry but Longitude present, assign here
    :param yCol: if no shapely geometry but Latitude present, assign here
    :returns: a geodataframe of the node objects in the graph
    """

    nodes = []
    keys = []

    if attr_list is None:
        for u, data in G.nodes(data = True):
            keys.append(list(data.keys()))
        flatten = lambda l: [item for sublist in l for item in sublist]
        attr_list = list(set(flatten(keys)))

    if geometry_tag in attr_list:
        non_geom_attr_list = attr_list
        non_geom_attr_list.remove(geometry_tag)
    else:
        non_geom_attr_list = attr_list

    z = 0

    for u, data in G.nodes(data=True):

        if geometry_tag not in attr_list and xCol in attr_list and yCol in attr_list :
            try:
                new_column_info = {
                'node_ID': u,
                'geometry': Point(data[xCol], data[yCol]),
                'x': data[xCol],
                'y': data[yCol]}
            except:
                print('Skipped due to missing geometry data:',(u, data))
        else:
            try:
                new_column_info = {
                'node_ID': u,
                'geometry': data[geometry_tag],
                'x':data[geometry_tag].x,
                'y':data[geometry_tag].y}
            except:
                print((u, data))

        for i in non_geom_attr_list:
            try:
                new_column_info[i] = data[i]
            except:
                pass

        nodes.append(new_column_info)
        z += 1

    nodes_df = pd.DataFrame(nodes)
    nodes_df = nodes_df[['node_ID',*non_geom_attr_list,geometry_tag]]
    nodes_df = nodes_df.drop_duplicates(subset=['node_ID'], keep='first')
    nodes_gdf = gpd.GeoDataFrame(nodes_df, geometry=nodes_df.geometry, crs = crs)

    return nodes_gdf

def edge_gdf_from_graph(G, crs = {'init' :'epsg:4326'}, attr_list = None, geometry_tag = 'geometry', xCol='x', yCol = 'y'):
    """
    Function for generating a GeoDataFrame from a networkx Graph object

    :param G: (required) a graph object G
    :param crs: (optional) projection of format {'init' :'epsg:4326'}. Defaults to WGS84. Note: here we are defining the crs of the input geometry -we do NOT reproject to this crs. To reproject, consider using geopandas' to_crs method on the returned gdf.
    :param attr_list: (optional) list of the keys which you want to be moved over to the GeoDataFrame.
    :param geometry_tag: (optional) the key in the data dictionary for each edge which contains the geometry info.
    :param xCol: (optional) if no geometry is present in the edge data dictionary, the function will try to construct a straight line between the start and end nodes, if geometry information is present in their data dictionaries.  Pass the Longitude info as 'xCol'.
    :param yCol: (optional) likewise, determining the Latitude tag for the node's data dictionary allows us to make a straight line geometry where an actual geometry is missing.
    :returns: a GeoDataFrame object of the edges in the graph
    """

    edges = []
    keys = []

    if attr_list is None:
        for u, v, data in G.edges(data = True):
            keys.append(list(data.keys()))
        flatten = lambda l: [item for sublist in l for item in sublist]
        keys = list(set(flatten(keys)))
        if geometry_tag in keys:
            keys.remove(geometry_tag)
        attr_list = keys

    for u, v, data in G.edges(data=True):

        if geometry_tag in data:
            # if it has a geometry attribute (a list of line segments), add them
            # to the list of lines to plot
            geom = data[geometry_tag]

        else:
            # if it doesn't have a geometry attribute, the edge is a straight
            # line from node to node
            x1 = G.nodes[u][xCol]
            y1 = G.nodes[u][yCol]
            x2 = G.nodes[v][xCol]
            y2 = G.nodes[v][yCol]
            geom = LineString([(x1, y1), (x2, y2)])

        new_column_info = {
            'stnode':u,
            'endnode':v,
            'geometry':geom}

        for i in attr_list:
            try:
                new_column_info[i] = data[i]
            except:
                pass

        edges.append(new_column_info)

    edges_df = pd.DataFrame(edges)
    edges_df = edges_df[['stnode','endnode',*attr_list,'geometry']]
    edges_gdf = gpd.GeoDataFrame(edges_df, geometry = edges_df.geometry, crs = crs)

    return edges_gdf

def graph_nodes_intersecting_polygon(G, polygons, crs = None):

    """
    Function for generating GeoDataFrame from Graph. Note: ensure any GeoDataFrames are in the same projection before using function, or pass a crs

    :param G: a Graph object OR node geodataframe
    :param crs: a crs object of form {'init':'epsg:XXXX'}. If passed, matches both inputs to this crs.
    :returns: a list of the nodes intersecting the polygons
    """

    if type(G) == nx.classes.multidigraph.MultiDiGraph:
        graph_gdf = node_gdf_from_graph(G)

    elif type(G) == gpd.geodataframe.GeoDataFrame:
        graph_gdf = G

    else:
        raise ValueError('Expecting a graph or node geodataframe for G!')

    if type(polygons) != gpd.geodataframe.GeoDataFrame:
        raise ValueError('Expecting a geodataframe for polygon(s)!')

    if crs != None and graph_gdf.crs != crs:
            graph_gdf = graph_gdf.to_crs(crs)

    if crs != None and polygons.crs != crs:
            polygons = polygons.to_crs(crs)

    if polygons.crs != graph_gdf.crs:
        raise ValueError('crs mismatch detected! aborting process')

    intersecting_nodes = []
    for poly in polygons.geometry:

        def chck(x, poly):
            if poly.contains(x):
                return 1
            else:
                return 0

        graph_gdf['intersecting'] = graph_gdf['geometry'].apply(lambda x: chck(x, poly))
        intersecting_nodes.append(list(graph_gdf['node_ID'].loc[graph_gdf['intersecting'] == 1]))

    intersecting_nodes = [j for i in intersecting_nodes for j in i]
    unique_intersecting_nodes = list(set(intersecting_nodes))
    return unique_intersecting_nodes

def graph_edges_intersecting_polygon(G, polygons, mode, crs = None, fast = True):
    """
    Function for identifying edges of a graph that intersect polygon(s). Ensure any GeoDataFrames are in the same projection before using function, or pass a crs.

    :param G: a Graph object
    :param polygons: a GeoDataFrame containing one or more polygons
    :param mode: a string, either 'contains' or 'intersecting'
    :param crs: If passed, will reproject both polygons and graph edge gdf to this projection.
    :param fast: (default: True): we can cheaply test whether an edge intersects a polygon gdf by checking whether either the start or end nodes are within a polygon. If both are, then we return 'contained'; if at least one is, we can return 'intersects'. If we set fast to False, then we iterate through each geometry one at a time, and check to see whether the geometry object literally intersects the polygon geodataframe, one at a time. May be computationally intensive!
    :returns: a list of the edges intersecting the polygons
    """

    if type(G) == nx.classes.multidigraph.MultiDiGraph:
        node_graph_gdf = node_gdf_from_graph(G)
        edge_graph_gdf = edge_gdf_from_graph(G)
    else:
        raise ValueError('Expecting a graph for G!')

    if type(polygons) != gpd.geodataframe.GeoDataFrame:
        raise ValueError('Expecting a geodataframe for polygon(s)!')

    if crs != None and node_graph_gdf.crs != crs:
            node_graph_gdf = node_graph_gdf.to_crs(crs)

    if crs != None and polygons.crs != crs:
            polygons = polygons.to_crs(crs)

    if polygons.crs != node_graph_gdf.crs:
        raise ValueError('crs mismatch detected! aborting process')

    intersecting_nodes = graph_nodes_intersecting_polygon(node_graph_gdf, polygons, crs)

    if fast == True:

        if mode == 'contains':
            edge_graph_gdf = edge_graph_gdf.loc[(edge_graph_gdf.stnode.isin(intersecting_nodes)) &
                                     (edge_graph_gdf.endnode.isin(intersecting_nodes))]
        elif mode == 'intersects':
            edge_graph_gdf = edge_graph_gdf.loc[(edge_graph_gdf.stnode.isin(intersecting_nodes)) |
                                     (edge_graph_gdf.endnode.isin(intersecting_nodes))]
    elif fast == False:
        poly = unary_union(polygons.geometry)

        if mode == 'contains':
            edge_graph_gdf = edge_graph_gdf.loc[(edge_graph_gdf.within(poly))]

        elif mode == 'intersects':
            edge_graph_gdf = edge_graph_gdf.loc[(edge_graph_gdf.intersects(poly))]

    else:
        raise ValueError("'fast' requires a boolean input!!")

    return edge_graph_gdf

def sample_raster(G, tif_path, property_name = 'RasterValue'):
    """
    Function for attaching raster values to corresponding graph nodes. Ensure any GeoDataFrames / graphs are in the same projection before using function, or pass a crs

    :param G: a graph containing one or more nodes
    :param tif_path: a raster or path to a tif
    :param property_name: a property name for the value of the raster attached to the node
    :returns: a graph
    """

    import rasterio

    if type(G) == nx.classes.multidigraph.MultiDiGraph or type(G) == nx.classes.digraph.DiGraph:
        pass
    else:
        raise ValueError('Expecting a graph or geodataframe for G!')

    # generate dictionary of {node ID: point} pairs
    try:
        list_of_nodes = {}
        for u, data in G.nodes(data=True):
            list_of_nodes.update({u:(data['x'], data['y'])})
    except:
        raise ValueError('loading point geometry went wrong. Ensure node data dict includes x, y values!')

    # load raster
    try:
        dataset = rasterio.open(os.path.join(tif_path))
    except:
        raise ValueError('Expecting a path to a .tif file!')

    # create list of values, throw out nodes that don't intersect the bounds of the raster
    b = dataset.bounds
    datasetBoundary = box(b[0], b[1], b[2], b[3])
    selKeys = []
    selPts = []
    for key, pt in list_of_nodes.items():
        if Point(pt[0], pt[1]).intersects(datasetBoundary):
            selPts.append(pt)
            selKeys.append(key)
    raster_values = list(dataset.sample(selPts))
    raster_values = [x[0] for x in raster_values]

    # generate new dictionary of {node ID: raster values}
    ref = dict(zip(selKeys, raster_values))

    # load new values onto node data dictionary
    missedCnt = 0
    for u, data in G.nodes(data=True):
        try:
            data[property_name] = ref[u]
        except:
            missedCnt += 1
            logging.info("Could not add raster value to node %s" % u)
    logging.info("Number of original nodes: %s" % len(G.nodes))
    logging.info("Number of missed nodes in raster: %d" % missedCnt)
    logging.info("Number of nodes that intersected raster: %d" % len(selKeys))

    return G

def generate_isochrones(G, origins, thresh, weight = None, stacking = False):
    """
    Function for generating isochrones from one or more graph nodes. Ensure any GeoDataFrames / graphs are in the same projection before using function, or pass a crs

    :param G: a graph containing one or more nodes
    :param orgins: a list of node IDs that the isochrones are to be generated from
    :param thresh: The time threshold for the calculation of the isochrone
    :param weight: Name of edge weighting for calculating 'distances'. For isochrones, should be time expressed in seconds. Defaults to time expressed in seconds.
    :param stacking: If True, returns number of origins that can be reached from that node. If false, max = 1
    :returns: The original graph with a new data property for the nodes and edges included in the isochrone
    """

    if type(origins) == list and len(origins) >= 1:
        pass
    else:
        raise ValueError('Ensure isochrone centers (origins object) is a list containing at least one node ID!')

    ddict = list(G.nodes(data = True))[:1][0][1]

    if weight == None:
        if 'time' not in ddict.keys():
            raise ValueError('need "time" key in edge value dictionary!')
        else:
            weight = 'time'

    sub_graphs = []
    for node in origins:
        sub_graphs.append(nx.ego_graph(G, node, thresh, distance = weight))

    reachable_nodes = []
    for graph in sub_graphs:
        reachable_nodes.append(list(graph.nodes))

    reachable_nodes = [j for i in reachable_nodes for j in i]

    if stacking == False:

        reachable_nodes = set(reachable_nodes)

        for u, data in G.nodes(data=True):
            if u in reachable_nodes:
                data[thresh] = 1
            else:
                data[thresh] = 0

    elif stacking == True:

        reachable_nodes = Counter(reachable_nodes)

        for u, data in G.nodes(data=True):
            if u in reachable_nodes:
                data[thresh] = reachable_nodes[u]
            else:
                data[thresh] = 0
    else:
        raise ValueError('stacking must either be True or False!')

    return G

def make_iso_polys(G, origins, trip_times, edge_buff=10, node_buff=25, infill=False, weight = 'time', measure_crs = {'init':'epsg:4326'}):
    """
    Function for adding a time value to edge dictionaries

    :param G: a graph object
    :param origins: a list object of node IDs from which to generate an isochrone poly object
    :param trip_times: a list object containing the isochrone values
    :param edge_buff: the thickness with which to buffer included edges
    :param node_buff: the thickness with whcih to buffer included nodes
    :param infill: If True, will remove any holes in isochrones
    :param weight: The edge weight to use when appraising travel times.
    :param crs: measurement crs, object of form {'init':'epsg:XXXX'}
    """

    default_crs = {'init':'epsg:4326'}

    if type(origins) == list and len(origins) >= 1:
        pass
    else:
        raise ValueError('Ensure isochrone centers ("origins" object) is a list containing at least one node ID!')

    isochrone_polys, nodez, tt = [], [], []

    for trip_time in sorted(trip_times, reverse=True):

        for _node_ in origins:

            subgraph = nx.ego_graph(G, _node_, radius = trip_time, distance = weight)
            node_points = [Point((data['x'], data['y'])) for node, data in subgraph.nodes(data=True)]

            if len(node_points) > 1:

                nodes_gdf = gpd.GeoDataFrame({'id': subgraph.nodes()}, geometry=node_points, crs = default_crs)
                nodes_gdf = nodes_gdf.set_index('id')

                edge_lines = []
                for n_fr, n_to in subgraph.edges():
                    f = nodes_gdf.loc[n_fr].geometry
                    t = nodes_gdf.loc[n_to].geometry
                    edge_lines.append(LineString([f,t]))

                edge_gdf = gpd.GeoDataFrame({'geoms':edge_lines}, geometry = 'geoms', crs = default_crs)

                if measure_crs != None and nodes_gdf.crs != measure_crs:
                    nodes_gdf = nodes_gdf.to_crs(measure_crs)
                    edge_gdf = edge_gdf.to_crs(measure_crs)

                n = nodes_gdf.buffer(node_buff).geometry
                e = edge_gdf.buffer(edge_buff).geometry

                all_gs = list(n) + list(e)

                new_iso = gpd.GeoSeries(all_gs).unary_union

                # If desired, try and "fill in" surrounded
                # areas so that shapes will appear solid and blocks
                # won't have white space inside of them

                if infill:
                    new_iso = Polygon(new_iso.exterior)

                isochrone_polys.append(new_iso)
                tt.append(trip_time)
                nodez.append(str(_node_))
            else:
                pass

    gdf = gpd.GeoDataFrame({'geometry':isochrone_polys,'thresh':tt,'nodez':_node_}, crs = measure_crs, geometry = 'geometry')
    gdf = gdf.to_crs(default_crs)

    return gdf

def find_hwy_distances_by_class(G, distance_tag='length'):
    """
    Function for finding out the different highway classes in the graph and their respective lengths

    :param G: a graph object
    :param distance_tag: specifies which edge attribute represents length
    :returns: a dictionary that has each class and the total distance per class
    """

    if type(G) == nx.classes.multidigraph.MultiDiGraph or type(G) == nx.classes.digraph.DiGraph:
        pass
    else:
        raise ValueError('Expecting a graph or geodataframe for G!')

    G_adj = G.copy()

    class_list = []

    for u, v, data in G_adj.edges(data=True):
        #print(data['highway'])
        if type(data['highway']) == list:
                if data['highway'][0] not in class_list:
                    class_list.append(data['highway'][0])
        else:
            if data['highway'] not in class_list:
                class_list.append(data['highway'])
    
    class_dict = { i : 0 for i in class_list }

    for i in class_list:
        for u, v, data in G_adj.edges(data=True):
            if type(data['highway']) == list:
                if data['highway'][0] == i:
                    class_dict[i] += data[distance_tag]
            else:
                if data['highway'] == i:
                    class_dict[i] += data[distance_tag]

    return class_dict

def find_graph_avg_speed(G, distance_tag, time_tag):
    """
    Function for finding the average speed per km for the graph. It will sum up the total meters in the graph and the total time (in sec). \
    Then it will convert m/sec to km/hr. This function needs the 'convert_network_to_time' function to have run previously.
    
    :param G:
      a graph containing one or more nodes
    :param distance_tag:
      the key in the dictionary for the field currently containing a distance in meters
    :param time_tag:
      time to traverse the edge in seconds
    :returns:
      The average speed for the whole graph in km per hr
    """

    if type(G) == nx.classes.multidigraph.MultiDiGraph or type(G) == nx.classes.digraph.DiGraph:
        pass
    else:
        raise ValueError('Expecting a graph or geodataframe for G!')

    G_adj = G.copy()

    total_meters = 0
    total_sec = 0

    for u, v, data in G_adj.edges(data=True):

        total_meters = total_meters + data[distance_tag]
        total_sec = total_sec + data[time_tag]

    # perform conversion
    # ex. 5m/1sec = .005/.00027 = 18.51 kph
    avg_speed_kmph = (total_meters/1000)/(total_sec/3600)

    return avg_speed_kmph

def convert_network_to_time(G, distance_tag, graph_type = 'drive', road_col = 'highway', speed_dict = None, walk_speed = 4.5, factor = 1, default = None):
    """
    Function for adding a time value to edge dictionaries. Ensure any GeoDataFrames / graphs are in the same projection before using function, or pass a crs.

    DEFAULT SPEEDS:

               speed_dict = {
               'residential': 20,  # kmph
               'primary': 40, # kmph
               'primary_link':35,
               'motorway':50,
               'motorway_link': 45,
               'trunk': 40,
               'trunk_link':35,
               'secondary': 30,
               'secondary_link':25,
               'tertiary':30,
               'tertiary_link': 25,
               'unclassified':20
               }

    :param G: a graph containing one or more nodes
    :param distance_tag: the key in the dictionary for the field currently
               containing a distance in meters
    :param road_col: key for the road type in the edge data dictionary
    :param graph_type: set to either 'drive' or 'walk'. IF walk - will set time = walking time across all segment, using the supplied walk_speed. IF drive - will use a speed dictionary for each road type, or defaults as per the note below.
    :param speed_dict: speed dictionary to use. If not supplied, reverts to
               defaults
    :param walk_speed: specify a walkspeed in km/h
    :param factor: allows you to scale up / down distances if saved in a unit other than meters. Set to 1000 if length in km.
    :param default: if highway type not in the speed_dict, use this road class as an in-fill value for time.
    :returns: The original graph with a new data property for the edges called 'time'
    """

    if type(G) == nx.classes.multidigraph.MultiDiGraph or type(G) == nx.classes.digraph.DiGraph:
        pass
    else:
        raise ValueError('Expecting a graph or geodataframe for G!')

    G_adj = G.copy()

    for u, v, data in G_adj.edges(data=True):

        orig_len = data[distance_tag] * factor

        # Note that this is a MultiDiGraph so there could
        # be multiple indices here, I naively assume this is not
        # the case
        data['length'] = orig_len

        # get appropriate speed limit
        if graph_type == 'walk':
            speed = walk_speed

        elif graph_type == 'drive':

            if speed_dict == None:
                speed_dict = {
                'residential': 20,  # kmph
                'primary': 40, # kmph
                'primary_link':35,
                'motorway':50,
                'motorway_link': 45,
                'trunk': 40,
                'trunk_link':35,
                'secondary': 30,
                'secondary_link':25,
                'tertiary':30,
                'tertiary_link': 25,
                'unclassified':20
                }

            highwayclass = data[road_col]

            if type(highwayclass) == list:
                highwayclass = highwayclass[0]

            if highwayclass in speed_dict.keys():
                speed = speed_dict[highwayclass]
            else:
                if default == None:
                    speed = 20
                else:
                    speed = speed_dict[default]

        else:
            raise ValueError('Expecting either a graph_type of "walk" or "drive"!')

        # perform conversion
        kmph = (orig_len / 1000) / speed
        in_seconds = kmph * 60 * 60
        data['time'] = in_seconds

        # And state the mode, too
        data['mode'] = graph_type

    return G_adj

def assign_traffic_times(G, mb_token, accepted_road_types = ['trunk','trunk_link','primary','primary_link','secondary','secondary_link','tertiary','tertiary_link','motorway','motorway_link'], verbose = False, road_col = 'infra_type', id_col = 'id'):
    """
    Function for querying travel times from the Mapbox "driving traffic" API. Queries are only made for the specified road types.

    :param G: a graph object of the road network
    :param mb_token: Mapbox token (retrieve from Mapbox account, starts with "pk:")
    :param road_types: a list of OSM road types for which to query traffic-aware travel time, defaults to main roads
    :param verbose: Set to true to monitor progress of queries and notify if any queries failed, defaults to False
    :param road_col: key for the road type in the edge data dictionary, defaults to 'infra_type'
    :param id_col: key for the id in the edge data dictionary, defaults to 'id'
    :returns: The original graph with two new data properties for the edges: 'mapbox_api' (a boolean set to True if the edge succesfuly received a trafic time value) and 'time_traffic' (travel time in seconds)
    """

    import json, time
    import urllib.request as url

    edges_all = edge_gdf_from_graph(G)

    def first_val(x):
      if isinstance(x, list):
        return x[0]
      else:
        return x

    edges_all[road_col] = edges_all[road_col].apply(lambda x: first_val(x))

    # print('print edges_all')
    # print(edges_all[road_col][390:400])

    print('print unique roads')
    # may not of orginally worked because some fields can contain multiple road tags in a list. Ex. [motorway, trunk]. need to do pre-processing
    print(edges_all[road_col].unique())

    print('print accepted_road_types')
    print(accepted_road_types)

    # pre-process the id_col to make sure it has only one value, sometimes the osmid column can contain a list
    edges_all[id_col] = edges_all[id_col].apply(lambda x: first_val(x))

    # specific rows can be selected by using .isin method on a series.
    edges = edges_all[edges_all[road_col].isin(accepted_road_types)].copy()

    base_url = 'https://api.mapbox.com/directions/v5/mapbox/driving-traffic/'
    end_url = f'?&access_token={mb_token}'
    numcalls = 0
    loop_count = 1

    function_start = time.time()
    start = time.time()

    for idx, row in edges.iterrows():

        # build request
        start_x = G.nodes[row.stnode]['x']
        start_y = G.nodes[row.stnode]['y']
        end_x = G.nodes[row.endnode]['x']
        end_y = G.nodes[row.endnode]['y']
        coordinates = str(start_x)+','+str(start_y)+';'+str(end_x)+','+str(end_y)
        request = base_url+coordinates+end_url
        r = url.urlopen(request)

        try:
            data = json.loads(r.read().decode('utf-8'))['routes'][0]['duration']
        except:
            data = np.nan

        # print(data)

        # assign response duration value to edges df
        edges.at[idx,'duration'] = data

        numcalls += 1
        if numcalls == 299:

            elapsed_seconds = (time.time() - start)%60
            # print('print elapsed_seconds without %: ' + str((time.time() - start)))
            # print('print elapsed_seconds: ' + str(elapsed_seconds))
            if verbose == True: print(f"Did {numcalls+1} calls in {elapsed_seconds:.2f} seconds, now wait {60-elapsed_seconds:.2f}, {(300*loop_count)/len(edges):.2%} complete")
            time.sleep(60-elapsed_seconds)

            # reset count
            numcalls = 0
            start = time.time()
            loop_count += 1

    edges['newID'] = edges['stnode'].astype(str)+"_"+edges['endnode'].astype(str)+"_"+edges[id_col].astype(str)
    edges_duration = edges[['newID','duration']].copy()
    edges_duration = edges_duration.set_index('newID')
    n_null = edges_duration.isnull().sum()['duration']

    if verbose == True and n_null > 0: print(f'query failed {n_null} times')

    edges_duration = edges_duration.dropna()

    for u, v, data in G.edges(data = True):
        newID = str(u) + "_" + str(v) + "_" + str(data[id_col])
        if newID in edges_duration.index:
            data['time_mapbox'] = edges_duration.loc[newID,'duration']
            data['mapbox_api'] = True
        else:
            data['mapbox_api'] = False

    print('complete function time: ' + str(time.time() - function_start))

    return G

def example_edge(G, n=1):
    """
    Prints out an example edge

    :param G: a graph object
    :param n: n - number of edges to print
    """
    i = list(G.edges(data = True))[:n]
    for j in i:
        print(j)

def example_node(G, n=1):
    """
    Prints out an example node

    :param G: a graph object
    :param n: number of nodes to print
    """

    i = list(G.nodes(data = True))[:n]
    for j in i:
        print(j)

def calculate_OD(G, origins, destinations, fail_value, weight = 'time', weighted_origins = False):
    """
    Function for generating an origin: destination matrix

    :param G: a graph containing one or more nodes
    :param fail_value: the value to return if the trip cannot be completed (implies some sort of disruption / disconnected nodes)
    :param origins: a list of the node IDs to treat as origins points
    :param destinations: a list of the node IDs to treat as destinations
    :param weight: use edge weight of 'time' unless otherwise specified
    :param weighted_origins: equals 'true' if the origins have weights. If so, the input to 'origins' must be dictionary instead of a list, where the keys are the origin IDs and the values are the weighted demands.
    :returns: a numpy matrix of format OD[o][d] = shortest time possible
    """
    
    if weighted_origins == True:
        print('weighted_origins equals true')
        OD = np.zeros((len(origins), len(destinations)))
        #dictionary key length
        o = 0
        #loop through dictionary
        for key,value in origins.items():
            origin = key
            for d in range(0,len(destinations)):
                destination = destinations[d]
                #find the shortest distance between the origin and destination
                distance = nx.dijkstra_path_length(G, origin, destination, weight = weight)
                # calculate weighted distance
                weighted_distance = distance * float(value)
                OD[o][d] = weighted_distance
            o += 1

    else:
        flip = 0
        if len(origins) > len(destinations):
            flip = 1
            o_2 = destinations
            destinations = origins
            origins = o_2

        #origins will be number or rows, destinations will be number of columns
        OD = np.zeros((len(origins), len(destinations)))

        for o in range(0, len(origins)):
            origin = origins[o]
            results_dict = nx.single_source_dijkstra_path_length(G, origin, cutoff = None, weight = weight)

            for d in range(0, len(destinations)):
                destination = destinations[d]
                if destination in results_dict.keys():
                    OD[o][d] = results_dict[destination]
                else:
                    OD[o][d] = fail_value

        if flip == 1:
            OD = np.transpose(OD)

    return OD

def disrupt_network(G, property, thresh, fail_value):
    """
    Function for disrupting a graph given a threshold value against a node's value. Any edges which bind to broken nodes have their 'time' property set to fail_value

    :param G: REQUIRED a graph containing one or more nodes and one or more edges
    :param property: the element in the data dictionary for the edges to test
    :param thresh: values of data[property] above this value are disrupted
    :param fail_value: The data['time'] property is set to this value to simulate the removal of the edge
    :returns: a modified graph with the edited 'time' attribute
    """
    G_copy = G.copy()

    broken_nodes = []

    for u, data in G_copy.nodes(data = True):

        if data[property] > thresh:

            broken_nodes.append(u)

    print('nodes disrupted: %s' % len(broken_nodes))
    i = 0
    for u, v, data in G_copy.edges(data = True):

        if u in broken_nodes or v in broken_nodes:

            data['time'] = fail_value
            i+=1

    print('edges disrupted: %s' % i)
    return G_copy

def randomly_disrupt_network(G, edge_frac, fail_value):
    """
    Function for randomly disurpting a network. NOTE: requires the graph to have an 'edge_id' value in the edge data dictionary. This DOES NOT have to be unique.

    :param G: a graph containing one or more nodes and one or more edges
    :param edge_frac: the percentage of edges to destroy. Integer rather than decimal, e.g. 5 = 5% of edges
    :param fail_value: the data['time'] property is set to this value to simulate the removal of the edge      
    :returns: a modified graph with the edited 'time' attribute the list of edge IDs randomly chosen for destruction
    """

    edgeid = []

    for u,v, data in G.edges(data = True):
        edgeid.append(data['edge_id'])

    num_to_destroy = math.floor(len(edgeid) / 2 * (edge_frac / 100))

    destroy_list = list(np.random.randint(low = 0, high = max(edgeid), size = [num_to_destroy]))

    G_adj = G.copy()

    for u, v, data in G_adj.edges(data = True):
        if data['edge_id'] in destroy_list:
            data['time'] = fail_value

    return G_adj, destroy_list

def gravity_demand(G, origins, destinations, weight, maxtrips = 100, dist_decay = 1, fail_value = 99999999999):
    """
    Function for generating a gravity-model based demand matrix. Note: 1 trip will always be returned between an origin and a destination, even if weighting would otherewise be 0.
    :param origins: a list of node IDs. Must be in G.
    :param destinations: a list of node IDs Must be in G.
    :param weight: the gravity weighting of the nodes in the model, e.g. population
    :param fail_value: the data['time'] property is set to this value to simulate the removal of the edge
    :param maxtrips: normalize the number of trips in the resultant function to this number of trip_times
    :param dist_decay: parameter controlling the aggresion of discounting based on distance
    :returns: a numpy array describing the demand between o and d in terms of number of trips
    """

    maxtrips = 100
    dist_decay = 1

    demand = np.zeros((len(origins), len(destinations)))

    shortest_time = Calculate_OD(G, origins, destinations, fail_value)

    for o in range(0, len(origins)):
        for d in range(0, len(destinations)):
            if origins == destinations and o == d:
                demand[o][d] = 0
            else:
                normalized_dist = shortest_time[o][d] / shortest_time.max()
                demand[o][d] = (
                (G.node[origins[o]][weight] *
                G.node[destinations[d]][weight]) *
                np.exp(-1 * dist_decay * normalized_dist)
                )

    demand = ((demand / demand.max()) * maxtrips)
    demand = np.ceil(demand).astype(int)
    return demand

def unbundle_geometry(c):
    """
    Function for unbundling complex geometric objects. Note: shapely MultiLineString objects quickly get complicated. They may not show up when you plot them in QGIS. This function aims to make a .csv 'plottable'

    :param c: any object. This helper function is usually applied in lambda format against a pandas / geopandas dataframe. The idea is to try to return more simple versions of complex geometries for LineString and MultiLineString type objects.
    :returns: an unbundled geometry value that can be plotted.
    """

    if type(c) == list:
        objs = []
        for i in c:
            if type(i) == str:
                J = loads(i)
                if type(J) == LineString:
                    objs.append(J)
                if type(J) == MultiLineString:
                    for j in J:
                        objs.append(j)
            elif type(i) == MultiLineString:
                for j in i:
                    objs.append(j)
            elif type(i) == LineString:
                objs.append(i)
            else:
                pass
            mls = MultiLineString(objs)
            ls = linemerge(mls)
        return ls
    elif type(c) == str:
        return loads(c)
    else:
        return c

def save(G, savename, wpath, pickle = True, edges = True, nodes = True):
    """
    function used to save a graph object in a variety of handy formats

    :param G: a graph object
    :param savename: the filename, WITHOUT extension
    :param wpath: the write path for where the user wants the files saved
    :param pickle: if set to false, will not save a pickle of the graph
    :param edges: if set to false, will not save an edge gdf
    :param nodes: if set to false, will not save a node gdf
    """

    if nodes == True:
        new_node_gdf = node_gdf_from_graph(G)
        new_node_gdf.to_csv(os.path.join(wpath, '%s_nodes.csv' % savename))
    if edges == True:
        new_edge_gdf = edge_gdf_from_graph(G)
        new_edge_gdf.to_csv(os.path.join(wpath, '%s_edges.csv' % savename))
    if pickle == True:
        nx.write_gpickle(G, os.path.join(wpath, '%s.pickle' % savename))

def add_missing_reflected_edges(G):
    """
    function for adding any missing reflected edges - makes all edges bidirectional. This is essential for routing with simplified graphs

    :param G: a graph object
    """
    unique_edges = []
    missing_edges = []

    for u, v in G.edges(data = False):
        unique_edges.append((u,v))
    for u, v, data in G.edges(data = True):
        if (v, u) not in unique_edges:
            missing_edges.append((v,u,data))
    G2 = G.copy()
    G2.add_edges_from(missing_edges)
    print(G2.number_of_edges())
    return G2

def remove_duplicate_edges(G, max_ratio = 1.5):
    """
    function for deleting duplicated edges - where there is more than one edge connecting a node pair. USE WITH CAUTION - will change both topological relationships and node maps
    :param G: a graph object
    :param max_ratio: most of the time we see duplicate edges that are clones of each other. Sometimes, however, there are valid duplicates. These occur if multiple roads connect two junctions uniquely and without interruption - e.g. two roads running either side of a lake which meet at either end. The idea here is that valid 'duplicate edges' will have geometries of materially different length. Hence, we include a ratio - defaulting to 1.5 - beyond which we are sure the duplicates are valid edges, and will not be deleted.
    """

    G2 = G.copy()
    uniques = []
    deletes = []
    for u, v, data in G2.edges(data = True):
        if (u,v) not in uniques:
            uniques.append((v,u))
            t = G2.number_of_edges(u, v)
            lengths = []
            for i in range(0,t):
                lengths.append(G2.edges[u,v,i]['length'])
            if max(lengths) / min(lengths) >= max_ratio:
                pass
            else:
                deletes.append((u,v))

    for d in deletes:
        G2.remove_edge(d[0],d[1])
    print(G2.number_of_edges())
    return G2

def convert_to_MultiDiGraph(G):
    """
    takes any graph object, loads it into a MultiDiGraph type Networkx object
    :param G: a graph object
    """
    a = nx.MultiDiGraph()

    node_bunch = []
    for u, data in G.nodes(data = True):
        node_bunch.append((u,data))

    a.add_nodes_from(node_bunch)

    edge_bunch = []
    for u, v, data in G.edges(data = True):
        if 'Wkt' in data.keys():
            data['Wkt'] = str(data['Wkt'])
        edge_bunch.append((u,v,data))

    a.add_edges_from(edge_bunch)
    return a

#### NETWORK SIMPLIFICATION ####

def simplify_junctions(G, measure_crs, in_crs = {'init': 'epsg:4326'}, thresh = 25):
    """
    simplifies topology of networks by simplifying node clusters into single nodes.

    :param G: a graph object
    :param measure_crs: the crs to make the measurements inself.
    :param in_crs: the current crs of the graph's geometry properties. By default, assumes WGS 84 (epsg 4326)
    :param thresh: the threshold distance in which to simplify junctions. By default, assumes 25 meters
    """

    G2 = G.copy()

    gdfnodes = node_gdf_from_graph(G2)
    gdfnodes_proj_buffer = gdfnodes.to_crs(measure_crs)
    gdfnodes_proj_buffer = gdfnodes_proj_buffer.buffer(thresh)
    juncs_gdf = gpd.GeoDataFrame(pd.DataFrame({'geometry':unary_union(gdfnodes_proj_buffer)}), crs = measure_crs, geometry = 'geometry')
    juncs_gdf['area'] = juncs_gdf.area

    juncs_gdf_2 = juncs_gdf.copy()
    juncs_gdf_2 = juncs_gdf_2.loc[juncs_gdf_2.area > int(juncs_gdf.area.min() + 1)]
    juncs_gdf = juncs_gdf_2
    juncs_gdf = juncs_gdf.reset_index()
    juncs_gdf['obj_ID'] = juncs_gdf.index
    juncs_gdf['obj_ID'] = 'new_obj_'+juncs_gdf['obj_ID'].astype(str)

    juncs_gdf_unproj = juncs_gdf.to_crs(in_crs)
    juncs_gdf_unproj['centroid'] = juncs_gdf_unproj.centroid
    juncs_gdf_bound = gpd.sjoin(juncs_gdf_unproj, gdfnodes, how='left', op='intersects', lsuffix='left', rsuffix='right')
    juncs_gdf_bound = juncs_gdf_bound[['obj_ID','centroid','node_ID']]

    node_map = juncs_gdf_bound[['obj_ID','node_ID']]
    node_map = node_map.set_index('node_ID')
    node_dict = node_map['obj_ID'].to_dict()
    nodes_to_be_destroyed = list(node_dict.keys())

    centroid_map = juncs_gdf_bound[['obj_ID','centroid']]
    centroid_map = centroid_map.set_index('obj_ID')
    centroid_dict = centroid_map['centroid'].to_dict()
    new_node_IDs = list(centroid_dict.keys())

    # Add the new centroids of the junction areas as new nodes
    new_nodes = []
    for i in new_node_IDs:
        new_nodes.append((i, {'x':centroid_dict[i].x, 'y':centroid_dict[i].y}))
    G2.add_nodes_from(new_nodes)

    # modify edges - delete those where both u and v are to be removed, edit the others
    edges_to_be_destroyed = []
    new_edges = []

    for u, v, data in G2.edges(data = True):

        if type(data['Wkt']) == LineString:
            l = data['Wkt']
        else:
            l = loads(data['Wkt'])

        line_to_be_edited = l.coords

        if u in nodes_to_be_destroyed and v in nodes_to_be_destroyed:
            if node_dict[u] == node_dict[v]:
                edges_to_be_destroyed.append((u,v))

            else:
                new_ID_u = node_dict[u]
                new_point_u = centroid_dict[new_ID_u]
                new_ID_v = node_dict[v]
                new_point_v = centroid_dict[new_ID_v]

                if len(line_to_be_edited) > 2:
                    data['Wkt'] = LineString([new_point_u, *line_to_be_edited[1:-1], new_point_v])
                else:
                    data['Wkt'] = LineString([new_point_u, new_point_v])
                data['Type'] = 'dual_destruction'

                new_edges.append((new_ID_u,new_ID_v,data))
                edges_to_be_destroyed.append((u,v))

        else:

            if u in nodes_to_be_destroyed:
                new_ID_u = node_dict[u]
                u = new_ID_u

                new_point = centroid_dict[new_ID_u]
                coords = [new_point, *line_to_be_edited[1:]]
                data['Wkt'] = LineString(coords)
                data['Type'] = 'origin_destruction'

                new_edges.append((new_ID_u,v,data))
                edges_to_be_destroyed.append((u,v))

            elif v in nodes_to_be_destroyed:
                new_ID_v = node_dict[v]
                v = new_ID_v

                new_point = centroid_dict[new_ID_v]
                coords = [*line_to_be_edited[:-1], new_point]
                data['Wkt'] = LineString(coords)
                data['Type'] = 'destination_destruction'

                new_edges.append((u,new_ID_v,data))
                edges_to_be_destroyed.append((u,v))

            else:
                data['Type'] = 'legitimate'
                pass

    # remove old edges that connected redundant nodes to each other / edges where geometry needed to be changed
    G2.remove_edges_from(edges_to_be_destroyed)

    # ... and add any corrected / new edges
    G2.add_edges_from(new_edges)

    # remove now redundant nodes
    G2.remove_nodes_from(nodes_to_be_destroyed)

    print(G2.number_of_edges())

    return G2

def custom_simplify(G, strict=True):
    """
    Simplify a graph's topology by removing all nodes that are not intersections or dead-ends. Create an edge directly between the end points that encapsulate them, but retain the geometry of the original edges, saved as attribute in new edge.

    :param G: networkx multidigraph
    :param bool strict: if False, allow nodes to be end points even if they fail all other rules but have edges with different OSM IDs
    :returns: networkx multidigraph
    """

    def get_paths_to_simplify(G, strict=True):

        """
        Create a list of all the paths to be simplified between endpoint nodes.

        The path is ordered from the first endpoint, through the interstitial nodes,
        to the second endpoint. If your street network is in a rural area with many
        interstitial nodes between true edge endpoints, you may want to increase
        your system's recursion limit to avoid recursion errors.

        Parameters
        ----------
        G : networkx multidigraph
        strict : bool
            if False, allow nodes to be end points even if they fail all other rules
            but have edges with different OSM IDs

        Returns
        -------
        paths_to_simplify : list
        """

        # first identify all the nodes that are endpoints
        start_time = time.time()
        endpoints = set([node for node in G.nodes() if is_endpoint(G, node, strict=strict)])

        start_time = time.time()
        paths_to_simplify = []

        # for each endpoint node, look at each of its successor nodes
        for node in endpoints:
            for successor in G.successors(node):
                if successor not in endpoints:
                    # if the successor is not an endpoint, build a path from the
                    # endpoint node to the next endpoint node
                    try:
                        path = build_path(G, successor, endpoints, path=[node, successor])
                        paths_to_simplify.append(path)
                    except RuntimeError:
                        # recursion errors occur if some connected component is a
                        # self-contained ring in which all nodes are not end points.
                        # could also occur in extremely long street segments (eg, in
                        # rural areas) with too many nodes between true endpoints.
                        # handle it by just ignoring that component and letting its
                        # topology remain intact (this should be a rare occurrence)
                        # RuntimeError is what Python <3.5 will throw, Py3.5+ throws
                        # RecursionError but it is a subtype of RuntimeError so it
                        # still gets handled
                        pass

        return paths_to_simplify

    def is_endpoint(G, node, strict=True):
        """
        Return True if the node is a "real" endpoint of an edge in the network, \
        otherwise False. OSM data includes lots of nodes that exist only as points \
        to help streets bend around curves. An end point is a node that either: \
        1) is its own neighbor, ie, it self-loops. \
        2) or, has no incoming edges or no outgoing edges, ie, all its incident \
            edges point inward or all its incident edges point outward. \
        3) or, it does not have exactly two neighbors and degree of 2 or 4. \
        4) or, if strict mode is false, if its edges have different OSM IDs. \

        Parameters
        ----------
        G : networkx multidigraph

        node : int
            the node to examine
        strict : bool
            if False, allow nodes to be end points even if they fail all other rules \
            but have edges with different OSM IDs

        Returns
        -------
        bool
        """

        neighbors = set(list(G.predecessors(node)) + list(G.successors(node)))
        n = len(neighbors)
        d = G.degree(node)

        if node in neighbors:
            # if the node appears in its list of neighbors, it self-loops. this is
            # always an endpoint.
            return 'node in neighbours'

        # if node has no incoming edges or no outgoing edges, it must be an endpoint
        #elif G.out_degree(node)==0 or G.in_degree(node)==0:
            #return 'no in or out'

        elif not (n==2 and (d==2 or d==4)):
            # else, if it does NOT have 2 neighbors AND either 2 or 4 directed
            # edges, it is an endpoint. either it has 1 or 3+ neighbors, in which
            # case it is a dead-end or an intersection of multiple streets or it has
            # 2 neighbors but 3 degree (indicating a change from oneway to twoway)
            # or more than 4 degree (indicating a parallel edge) and thus is an
            # endpoint
            return 'condition 3'

        elif not strict:
            # non-strict mode
            osmids = []

            # add all the edge OSM IDs for incoming edges
            for u in G.predecessors(node):
                for key in G[u][node]:
                    osmids.append(G.edges[u, node, key]['osmid'])

            # add all the edge OSM IDs for outgoing edges
            for v in G.successors(node):
                for key in G[node][v]:
                    osmids.append(G.edges[node, v, key]['osmid'])

            # if there is more than 1 OSM ID in the list of edge OSM IDs then it is
            # an endpoint, if not, it isn't
            return len(set(osmids)) > 1

        else:
            # if none of the preceding rules returned true, then it is not an endpoint
            return False

    def build_path(G, node, endpoints, path):
        """
        Recursively build a path of nodes until you hit an endpoint node.

        :param G: networkx multidigraph
        :param int node: the current node to start from
        :param set endpoints: the set of all nodes in the graph that are endpoints
        :param list path: the list of nodes in order in the path so far
        :returns list: paths_to_simplify
        """

        # for each successor in the passed-in node
        for successor in G.successors(node):
            if successor not in path:
                # if this successor is already in the path, ignore it, otherwise add
                # it to the path
                path.append(successor)
                if successor not in endpoints:
                    # if this successor is not an endpoint, recursively call
                    # build_path until you find an endpoint
                    path = build_path(G, successor, endpoints, path)
                else:
                    # if this successor is an endpoint, we've completed the path,
                    # so return it
                    return path

        if (path[-1] not in endpoints) and (path[0] in G.successors(path[-1])):
            # if the end of the path is not actually an endpoint and the path's
            # first node is a successor of the path's final node, then this is
            # actually a self loop, so add path's first node to end of path to
            # close it
            path.append(path[0])

        return path

    ## MAIN PROCESS FOR CUSTOM SIMPLIFY ##

    G = G.copy()

    if type(G) != nx.classes.multidigraph.MultiDiGraph:
        G = ConvertToMultiDiGraph(G)
    initial_node_count = len(list(G.nodes()))
    initial_edge_count = len(list(G.edges()))
    all_nodes_to_remove = []
    all_edges_to_add = []

    # construct a list of all the paths that need to be simplified
    paths = get_paths_to_simplify(G, strict=strict)

    start_time = time.time()
    for path in paths:

        # add the interstitial edges we're removing to a list so we can retain
        # their spatial geometry
        edge_attributes = {}
        for u, v in zip(path[:-1], path[1:]):

            # there shouldn't be multiple edges between interstitial nodes
            if not G.number_of_edges(u, v) == 1:
                pass
            # the only element in this list as long as above check is True
            # (MultiGraphs use keys (the 0 here), indexed with ints from 0 and
            # up)
            edge = G.edges[u, v, 0]
            for key in edge:
                if key in edge_attributes:
                    # if this key already exists in the dict, append it to the
                    # value list
                    edge_attributes[key].append(edge[key])
                else:
                    # if this key doesn't already exist, set the value to a list
                    # containing the one value
                    edge_attributes[key] = [edge[key]]

        for key in edge_attributes:
            # don't touch the length attribute, we'll sum it at the end
            if key == 'Wkt':
                edge_attributes['Wkt'] = list(edge_attributes['Wkt'])
            elif key != 'length' and key != 'Wkt':      # if len(set(edge_attributes[key])) == 1 and not key == 'length':
                # if there's only 1 unique value in this attribute list,
                # consolidate it to the single value (the zero-th)
                edge_attributes[key] = edge_attributes[key][0]
            elif not key == 'length':
                # otherwise, if there are multiple values, keep one of each value
                edge_attributes[key] = list(set(edge_attributes[key]))

        # construct the geometry and sum the lengths of the segments
        edge_attributes['geometry'] = LineString([Point((G.nodes[node]['x'], G.nodes[node]['y'])) for node in path])
        edge_attributes['length'] = sum(edge_attributes['length'])

        # add the nodes and edges to their lists for processing at the end
        all_nodes_to_remove.extend(path[1:-1])
        all_edges_to_add.append({'origin':path[0],
                                 'destination':path[-1],
                                 'attr_dict':edge_attributes})

    # for each edge to add in the list we assembled, create a new edge between
    # the origin and destination
    for edge in all_edges_to_add:
        G.add_edge(edge['origin'], edge['destination'], **edge['attr_dict'])

    # finally remove all the interstitial nodes between the new edges
    G.remove_nodes_from(set(all_nodes_to_remove))

    msg = 'Simplified graph (from {:,} to {:,} nodes and from {:,} to {:,} edges) in {:,.2f} seconds'
    return G

def salt_long_lines(G, source, target, thresh = 5000, factor = 1, attr_list = None):
    print('WARNING: "factor behavior has changed! now divides rather than multiplies. This change brings gn.salt_long_lines into line with gn.convert_network_to_time" ')
    """
    Adds in new nodes to edges greater than a given length

    :param G: a graph object
    :param source: crs object in format 'epsg:4326'
    :param target: crs object in format 'epsg:32638'
    :param thresh: distance in metres after which to break edges.
    :param factor: edge lengths can be returned in units other than metres by specifying a numerical multiplication factor
    :param attr_dict: list of attributes to be saved onto new edges.
    """

    def cut(line, distance):
        # Cuts a line in two at a distance from its starting point
        if distance <= 0.0 or distance >= line.length:
            return [LineString(line)]

        coords = list(line.coords)
        for i, p in enumerate(coords):
            pd = line.project(Point(p))
            if pd == distance:
                return [LineString(coords[:i+1]),LineString(coords[i:])]
            if pd > distance:
                cp = line.interpolate(distance)
                return [LineString(coords[:i] + [(cp.x, cp.y)]),LineString([(cp.x, cp.y)] + coords[i:])]

    G2 = G.copy()

    # define transforms for exchanging between source and target projections

    project_WGS_UTM = partial(
                pyproj.transform,
                pyproj.Proj(init=source),
                pyproj.Proj(init=target))

    project_UTM_WGS = partial(
                pyproj.transform,
                pyproj.Proj(init=target),
                pyproj.Proj(init=source))

    long_edges, long_edge_IDs, unique_long_edges, new_nodes, new_edges = [], [], [], [], []

    # Identify long edges
    for u, v, data in G2.edges(data = True):

        # load geometry
        if type(data['Wkt']) == str:
            WGS_geom = loads(data['Wkt'])
        else:
            WGS_geom = unbundle_geometry(data['Wkt'])
        UTM_geom = transform(project_WGS_UTM, WGS_geom)

        # test geomtry length
        if UTM_geom.length > thresh:
            long_edges.append((u, v, data))
            long_edge_IDs.append((u,v))
            if (v, u) in long_edge_IDs:
                pass
            else:
                unique_long_edges.append((u, v, data))

    print('Identified %d unique edge(s) longer than %d. \nBeginning new node creation...' % (len(unique_long_edges), thresh))

    # iterate through one long edge for each bidirectional long edge pair

    j,o = 1, 0

    for u, v, data in unique_long_edges:

        # load geometry of long edge
        if type(data['Wkt']) == str:
            WGS_geom = loads(data['Wkt'])
        else:
            WGS_geom = unbundle_geometry(data['Wkt'])

        if WGS_geom.type == 'MultiLineString':
            WGS_geom = linemerge(WGS_geom)

        UTM_geom = transform(project_WGS_UTM, WGS_geom)

        # flip u and v if Linestring running from v to u, coordinate-wise
        u_x_cond = round(WGS_geom.coords[0][0], 6) == round(G.nodes()[u]['x'], 6)
        u_y_cond = round(WGS_geom.coords[0][1], 6) == round(G.nodes()[u]['y'], 6)

        v_x_cond = round(WGS_geom.coords[0][0], 6) == round(G.nodes()[v]['x'], 6)
        v_y_cond = round(WGS_geom.coords[0][1], 6) == round(G.nodes()[v]['y'], 6)

        if u_x_cond and u_y_cond:
            pass
        elif v_x_cond and v_y_cond:
            u, v = v, u
        else:
            print('ERROR!')

        # calculate number of new nodes to add along length
        number_of_new_points = UTM_geom.length / thresh

        # for each new node
        for i in range(0, int(number_of_new_points+1)):

            ## GENERATE NEW NODES ##

            cur_dist = (thresh * (i+1))

            # generate new geometry along line
            new_point = UTM_geom.interpolate(cur_dist)

            new_point_WGS = transform(project_UTM_WGS, new_point)

            node_data = {'geometry': new_point_WGS,
                        'x' : new_point_WGS.x,
                        'y': new_point_WGS.y}

            new_node_ID = str(u)+'_'+str(i+j)+'_'+str(o)

            # generate a new node as long as we aren't the final node
            if i < int(number_of_new_points):
                new_nodes.append((new_node_ID, node_data))

            ## GENERATE NEW EDGES ##
            try:
                # define geometry to be cutting (iterative)
                if i == 0:
                    geom_to_split = UTM_geom

                else:
                    geom_to_split = result[1]

                # cut geometry. result[0] is the section cut off, result[1] is remainder
                result = cut(geom_to_split, (thresh))

                t_geom = transform(project_UTM_WGS, result[0])

                edge_data = {'Wkt' : t_geom,
                            'length' : (int(result[0].length) / factor),
                            }

                if attr_list != None:
                    for attr in attr_list:
                        edge_data[attr] = data[attr]

                if i == 0:
                    prev_node_ID = u

                if i == int(number_of_new_points):
                    new_node_ID = v

                # append resulting edges to a list of new edges, bidirectional.
                new_edges.append((prev_node_ID,new_node_ID,edge_data))
                new_edges.append((new_node_ID,prev_node_ID,edge_data))

                o += 1

                prev_node_ID = new_node_ID
            except:
                pass

        j+=1

    # add new nodes and edges
    G2.add_nodes_from(new_nodes)
    G2.add_edges_from(new_edges)

    # remove the too-long edges
    for d in long_edges:
        G2.remove_edge(d[0],d[1])

    print('%d new edges added and %d removed to bring total edges to %d' % (len(new_edges),len(long_edges),G2.number_of_edges()))
    print('%d new nodes added to bring total nodes to %d' % (len(new_nodes),G2.number_of_nodes()))

    return G2

def pandana_snap(G, point_gdf, source_crs = 'epsg:4326', target_crs = 'epsg:4326', 
                    add_dist_to_node_col = True):
    """
    snaps points to a graph at very high speed
    :param G: a graph object
    :param point_gdf: a geodataframe of points, in the same source crs as the geometry of the graph object
    :param source_crs: crs object in format 'epsg:32638'
    :param target_crs: crs object in format 'epsg:32638'
    :param add_dist_to_node_col: return distance in metres to nearest node
    """

    in_df = point_gdf.copy()
    node_gdf = node_gdf_from_graph(G)

    if add_dist_to_node_col is True:

        project_WGS_UTM = partial(
                    pyproj.transform,
                    pyproj.Proj(init=source_crs),
                    pyproj.Proj(init=target_crs))

        in_df['Proj_geometry'] = in_df.apply(lambda x: transform(project_WGS_UTM, x.geometry), axis = 1)
        in_df = in_df.set_geometry('Proj_geometry')
        in_df['x'] = in_df.Proj_geometry.x
        in_df['y'] = in_df.Proj_geometry.y

        node_gdf['Proj_geometry'] = node_gdf.apply(lambda x: transform(project_WGS_UTM, x.geometry), axis = 1)
        node_gdf = node_gdf.set_geometry('Proj_geometry')
        node_gdf['x'] = node_gdf.Proj_geometry.x
        node_gdf['y'] = node_gdf.Proj_geometry.y

        G_tree = spatial.KDTree(node_gdf[['x','y']].as_matrix())

        distances, indices = G_tree.query(in_df[['x','y']].as_matrix())

        in_df['NN'] = list(node_gdf['node_ID'].iloc[indices])
        in_df['NN_dist'] = distances
        in_df = in_df.drop(['x','y','Proj_geometry'], axis = 1)

    else:
        in_df['x'] = in_df.geometry.x
        in_df['y'] = in_df.geometry.y
        G_tree = spatial.KDTree(node_gdf[['x','y']].as_matrix())

        distances, indices = G_tree.query(in_df[['x','y']].as_matrix())

        in_df['NN'] = list(node_gdf['node_ID'].iloc[indices])

    return in_df

def pandana_snap_single_point(G, shapely_point, source_crs = 'epsg:4326', target_crs = 'epsg:4326'):
    """
    snaps a point to a graph at very high speed

    :param G: a graph object
    :param shapely_point: a shapely point (ex. Point(x, y)), in the same source crs as the geometry of the graph object
    :param source_crs: crs object in format 'epsg:32638'
    :param target_crs: crs object in format 'epsg:32638'
    :param add_dist_to_node_col: return distance in metres to nearest node
    """

    node_gdf = node_gdf_from_graph(G)

    G_tree = spatial.KDTree(node_gdf[['x','y']].as_matrix())
    distances, indices = G_tree.query([[shapely_point.x,shapely_point.y]])

    #print("print distances, indices")
    #print(distances, indices)

    return_list = list(node_gdf['node_ID'].iloc[indices])

    return return_list[0]

def pandana_snap_points(source_gdf, target_gdf, source_crs = 'epsg:4326', target_crs = 'epsg:4326', add_dist_to_node_col = True):
    """
    snaps points to another GeoDataFrame at very high speed

    :param source_gdf: a geodataframe of points, in the same source crs as the geometry of the target_gdf
    :param target_gdf: a geodataframe of points, in the same source crs as the geometry of the source_gdf
    :param source_gdf: a geodataframe of points, in the same source crs as the geometry of the source_gdf
    :param source_crs: crs object in format 'epsg:32638'
    :param target_crs: crs object in format 'epsg:32638'
    :param add_dist_to_node_col: return distance in metres to nearest node
    """

    source_gdf = source_gdf.copy()
    target_gdf = target_gdf.copy()
    target_gdf['ID'] = target_gdf.index

    if add_dist_to_node_col is True:

        project_WGS_UTM = partial(
                    pyproj.transform,
                    pyproj.Proj(init=source_crs),
                    pyproj.Proj(init=target_crs))

        target_gdf['P'] = target_gdf.apply(lambda x: transform(project_WGS_UTM, x.geometry), axis = 1)
        target_gdf = target_gdf.set_geometry('P')
        target_gdf['x'] = target_gdf.P.x
        target_gdf['y'] = target_gdf.P.y

        source_gdf['P'] = source_gdf.apply(lambda x: transform(project_WGS_UTM, x.geometry), axis = 1)
        source_gdf = source_gdf.set_geometry('P')
        source_gdf['x'] = source_gdf.P.x
        source_gdf['y'] = source_gdf.P.y

        G_tree = spatial.KDTree(target_gdf[['x','y']].as_matrix())

        distances, indices = G_tree.query(source_gdf[['x','y']].as_matrix())

        source_gdf['NN'] = list(target_gdf['ID'].iloc[indices])

        source_gdf['NN_dist'] = distances

        source_gdf = source_gdf.drop(['x','y','P'], axis = 1)

    else:

        target_gdf['x'] = target_gdf.geometry.x
        target_gdf['y'] = target_gdf.geometry.y

        G_tree = spatial.KDTree(target_gdf[['x','y']].as_matrix())

        distances, indices = G_tree.query(source_gdf[['x','y']].as_matrix())

        source_gdf['NN'] = list(target_gdf['ID'].iloc[indices])

    return source_gdf

def join_networks(base_net, new_net, measure_crs, thresh = 500):
    """
    joins two networks together within a binding threshold

    :param base_net: a base network object (nx.MultiDiGraph)
    :param new_net: the network to add on to the base (nx.MultiDiGraph)
    :param measure_crs: the crs number of the measurement (epsg code)
    :param thresh: binding threshold - unit of the crs - default 500m
    """
    G_copy = base_net.copy()
    join_nodes_df = pandana_snap(G_copy,
                         node_gdf_from_graph(new_net),
                         source_crs = 'epsg:4326',
                         target_crs = 'epsg:%s' % measure_crs,
                         add_dist_to_node_col = True)

    join_nodes_df = join_nodes_df.sort_values(by = 'NN_dist', ascending = True)
    join_nodes_df = join_nodes_df.loc[join_nodes_df.NN_dist < thresh]

    nodes_to_add, edges_to_add = [],[]

    for u, data in new_net.nodes(data = True):
        u = 'add_net_%s' % u
        nodes_to_add.append((u,data))

    for u,v, data in new_net.edges(data = True):
        u = 'add_net_%s' % u
        v = 'add_net_%s' % v
        edges_to_add.append((u,v,data))

    gdf_base = node_gdf_from_graph(base_net)
    gdf_base = gdf_base.set_index('node_ID')

    for index, row in join_nodes_df.iterrows():
        u = 'add_net_%s' % row.node_ID
        v = row.NN
        data = {}
        data['length'] = row.NN_dist / 1000
        data['infra_type'] = 'net_glue'
        data['Wkt'] = LineString([row.geometry, gdf_base.geometry.loc[v]])
        edges_to_add.append((u, v, data))
        edges_to_add.append((v, u, data))

    G_copy.add_nodes_from(nodes_to_add)
    G_copy.add_edges_from(edges_to_add)

    G_copy = nx.convert_node_labels_to_integers(G_copy)

    return G_copy

def clip(G, bound, source_crs = 'epsg:4326', target_crs = 'epsg:4326', geom_col = 'geometry', largest_G = True):
    """
    Removes any edges that fall beyond a polygon, and shortens any other edges that do so

    :param G: a graph object.
    :param bound: a shapely polygon object
    :param source_crs: crs object in format 'epsg:4326'
    :param target_crs: crs object in format 'epsg:4326'
    :param geom_col: label name for geometry object
    :param largest_G: if True, takes largest remaining subgraph of G as G
    """

    edges_to_add, nodes_to_add = [],[]
    edges_to_remove, nodes_to_remove = [],[]

    if type(bound) == shapely.geometry.multipolygon.MultiPolygon or type(bound) == shapely.geometry.polygon.Polygon:
        pass
    else:
        raise ValueError('Bound input must be a Shapely Polygon or MultiPolygon object!')

    if type(G) != networkx.classes.multidigraph.MultiDiGraph:
        raise ValueError('Graph object must be of type networkx.classes.multidigraph.MultiDiGraph!')

    project_WGS_UTM = partial(
        pyproj.transform,
        pyproj.Proj(init=source_crs),
        pyproj.Proj(init=target_crs))

    G_copy = G.copy()
    print('pre_clip | nodes: %s | edges: %s' % (G_copy.number_of_nodes(), G_copy.number_of_edges()))

    existing_legitimate_point_geometries = {}
    for u, data in G_copy.nodes(data = True):
        geo_point = Point(round(data['x'],10),round(data['y'],10))
        if bound.contains(geo_point):
            existing_legitimate_point_geometries[u] = geo_point
        else:
            nodes_to_remove.append(u)

    iterator = 0
    done_edges = []

    for u, v, data in G_copy.edges(data = True):

        done_edges.append((v,u))

        if (u,v) in done_edges:
            pass

        else:
            # define basics from data dictionary
            infra_type = data['infra_type']
            #extract the geometry of the geom_col, if there is no explicit geometry, load the wkt
            try:
                geom = data[geom_col]
            except:
                geom = loads(data['Wkt'])

            # road fully within country - do nothing
            if bound.contains(geom) == True:
                pass

            # road fully outside country - remove entirely
            elif bound.intersects(geom) == False:

                edges_to_remove.append((u, v))
                edges_to_remove.append((v, u))
                nodes_to_remove.append(u)
                nodes_to_remove.append(v)

            # road partially in, partially out
            else:

                # start by removing existing edges
                edges_to_remove.append((u, v))
                edges_to_remove.append((v, u))

                # identify the new line sections inside the boundary
                new_geom = bound.intersection(geom)
                if type(new_geom) == shapely.geometry.multilinestring.MultiLineString:
                    new_geom = linemerge(new_geom)

                # If there is only one:
                if type(new_geom) == shapely.geometry.linestring.LineString:

                    new_nodes, new_edges, new_node_dict_entries, iterator = new_edge_generator(new_geom,infra_type,iterator,existing_legitimate_point_geometries,geom_col,project_WGS_UTM)
                    existing_legitimate_point_geometries.update(new_node_dict_entries)
                    nodes_to_add.append(new_nodes)
                    edges_to_add.append(new_edges)

                elif type(new_geom) == shapely.geometry.multilinestring.MultiLineString:

                    for n in new_geom:
                        new_nodes, new_edges, new_node_dict_entries, iterator = new_edge_generator(n,infra_type,iterator,existing_legitimate_point_geometries,geom_col, project_WGS_UTM)
                        existing_legitimate_point_geometries.update(new_node_dict_entries)
                        nodes_to_add.append(new_nodes)
                        edges_to_add.append(new_edges)

    # Remove bad geometries
    G_copy.remove_nodes_from(nodes_to_remove)
    G_copy.remove_edges_from(edges_to_remove)

    # Add new geometries
    nodes_to_add = [item for sublist in nodes_to_add for item in sublist]
    edges_to_add = [item for sublist in edges_to_add for item in sublist]
    G_copy.add_nodes_from(nodes_to_add)
    G_copy.add_edges_from(edges_to_add)

    # Re-label nodes
    G_copy = nx.convert_node_labels_to_integers(G_copy)
    print('post_clip | nodes: %s | edges: %s' % (G_copy.number_of_nodes(), G_copy.number_of_edges()))

    # Select only largest remaining graph
    if largest_G == True:
        list_of_Gs = list((nx.strongly_connected_component_subgraphs(G_copy)))
        list_length = list(len(i) for i in list_of_Gs)
        m = max(list_length)
        t = [i for i, j in enumerate(list_length) if j == m][0]
        max_G = list_of_Gs[t]
        G_copy = max_G

    return G_copy

def new_edge_generator(passed_geom, infra_type, iterator, existing_legitimate_point_geometries, geom_col, project_WGS_UTM):
    """
    Generates new edge and node geometries based on a passed geometry. WARNING: This is a child process of clip(), and shouldn't be run on its own

    :param passed_geom: a shapely Linestring object
    :param infra_type: the road / highway class of the passed geometry
    :param iterator: helps count the new node IDs to keep unique nodes
    :param existing_legitimate_point_geometries: a dictionary of points already created / valid in [u:geom] format
    :param project_WGS_UTM: projection object to transform passed geometries
    :param geom_col: label name for geometry object
    """

    edges_to_add = []
    nodes_to_add = []

    # new start and end points will be start and end of line
    u_geo = passed_geom.coords[0]
    v_geo = passed_geom.coords[-1]
    u_geom, v_geom = Point(round(u_geo[0],10),round(u_geo[1],10)), Point(round(v_geo[0],10),round(v_geo[1],10))

    # check to see if geometry already exists. If yes, assign u and v node IDs accordingly
    # else, make a new u and v ID
    if u_geom in existing_legitimate_point_geometries.values():
        u = list(existing_legitimate_point_geometries.keys())[list(existing_legitimate_point_geometries.values()).index(u_geom)]

    else:
        u = 'new_node_%s' % iterator
        node_data = {}
        node_data['x'] = u_geom.x
        node_data['y'] = u_geom.y
        nodes_to_add.append((u,node_data))
        iterator += 1

    if v_geom in existing_legitimate_point_geometries.values():
        v = list(existing_legitimate_point_geometries.keys())[list(existing_legitimate_point_geometries.values()).index(v_geom)]

    else:
        v = 'new_node_%s' % iterator
        node_data = {}
        node_data['x'] = v_geom.x
        node_data['y'] = v_geom.y
        nodes_to_add.append((v,node_data))
        iterator += 1

    # update the data dicionary for the new geometry
    UTM_geom = transform(project_WGS_UTM, passed_geom)
    edge_data = {}
    edge_data[geom_col] = passed_geom
    edge_data['length'] = UTM_geom.length / 1000
    edge_data['infra_type'] = infra_type

    # assign new edges to network
    edges_to_add.append((u, v, edge_data))
    edges_to_add.append((v, u, edge_data))

    # new node dict entries - add newly created geometries to library of valid nodes
    new_node_dict_entries = []

    for u, data in nodes_to_add:
        new_node_dict_entries.append((u,Point(round(data['x'],10),round(data['y'],10))))

    return nodes_to_add, edges_to_add, new_node_dict_entries, iterator

def reproject_graph(input_net, source_crs, target_crs):
    """
    Converts the node coordinates of a graph. Assumes that there are straight lines between the start and end nodes.

    :param input_net: a base network object (nx.MultiDiGraph)
    :param source_crs: The projection of the input_net (epsg code)
    :param target_crs: The projection input_net will be converted to (epsg code)
    """

    import networkx as nx
    import geopandas as gpd
    from shapely.geometry import Point
    from scipy import spatial
    from functools import partial
    import pyproj
    from shapely.ops import transform

    project_WGS_UTM = partial(
                pyproj.transform,
                pyproj.Proj(init=source_crs),
                pyproj.Proj(init=target_crs))

    i = list(input_net.nodes(data = True))

    for j in i:
        #print(j[1])
        #print(j[1]['x'])
        #print(transform(project_WGS_UTM,j[1]['geom']))
        j[1]['x'] = transform(project_WGS_UTM,j[1]['geom']).x
        j[1]['y'] = transform(project_WGS_UTM,j[1]['geom']).y
        j[1]['geom'] = transform(project_WGS_UTM,j[1]['geom'])


    return input_net

def euclidean_distance(lat1, lon1, lat2, lon2):
    """
    Calculate the great circle distance between two points on the earth (specified in decimal degrees)

    :param lat1: lat1
    :param lon1: lon1
    :param lat2: lat2
    :param lon2: lon2

    """

    from math import radians, cos, sin, asin, sqrt

    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a))
    km = 6367 * c
    return km




'''
source: https://gist.github.com/Tofull/49fbb9f3661e376d2fe08c2e9d64320e
'''

## Modules
# Elementary modules
from math import radians, cos, sin, asin, sqrt
import copy

# Graph module
import networkx

# Specific modules
import xml.sax # parse osm file
from pathlib import Path # manage cached tiles


def haversine(lon1, lat1, lon2, lat2, unit_m = True):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    default unit : km
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a))
    r = 6371 # Radius of earth in kilometers. Use 3956 for miles
    if (unit_m):
        r *= 1000
    return c * r


def download_osm(left, bottom, right, top, proxy = False, proxyHost = "10.0.4.2", proxyPort = "3128", cache = False, cacheTempDir = "/tmp/tmpOSM/", verbose = True):
    """ Return a filehandle to the downloaded data from osm api."""

    import urllib.request # To request the web


    if (cache):
        ## cached tile filename
        cachedTileFilename = "osm_map_{:.8f}_{:.8f}_{:.8f}_{:.8f}.map".format(left, bottom, right, top)

        if (verbose):
            print("Cached tile filename :", cachedTileFilename)

        Path(cacheTempDir).mkdir(parents = True, exist_ok = True) ## Create cache path if not exists

        osmFile = Path(cacheTempDir + cachedTileFilename).resolve() ## Replace the relative cache folder path to absolute path

        if osmFile.is_file():
            # download from the cache folder
            if (verbose):
                print("Tile loaded from the cache folder.")

            fp = urllib.request.urlopen("file://"+str(osmFile))
            return fp

    if (proxy):
        # configure the urllib request with the proxy
        proxy_handler = urllib.request.ProxyHandler({'https': 'https://' + proxyHost + ":" + proxyPort, 'http': 'http://' + proxyHost + ":" + proxyPort})
        opener = urllib.request.build_opener(proxy_handler)
        urllib.request.install_opener(opener)


    request = "http://api.openstreetmap.org/api/0.6/map?bbox=%f,%f,%f,%f"%(left,bottom,right,top)

    if (verbose):
        print("Download the tile from osm web api ... in progress")
        print("Request :", request)

    fp = urllib.request.urlopen(request)

    if (verbose):
        print("OSM Tile downloaded")

    if (cache):
        if (verbose):
            print("Write osm tile in the cache"
            )
        content = fp.read()
        with open(osmFile, 'wb') as f:
            f.write(content)

        if osmFile.is_file():
            if (verbose):
                print("OSM tile written in the cache")

            fp = urllib.request.urlopen("file://"+str(osmFile)) ## Reload the osm tile from the cache (because fp.read moved the cursor)
            return fp

    return fp


def read_osm(filename_or_stream, only_roads=True):
    """Read graph in OSM format from file specified by name or by stream object.
    Parameters
    ----------
    filename_or_stream : filename or stream object
    Returns
    -------
    G : Graph
    Examples
    --------
    >>> G=nx.read_osm(nx.download_osm(-122.33,47.60,-122.31,47.61))
    >>> import matplotlib.pyplot as plt
    >>> plt.plot([G.node[n]['lat']for n in G], [G.node[n]['lon'] for n in G], 'o', color='k')
    >>> plt.show()
    """
    osm = OSM(filename_or_stream)
    G = networkx.DiGraph()

    ## Add ways
    for w in osm.ways.values():
        if only_roads and 'highway' not in w.tags:
            continue

        if ('oneway' in w.tags):
            if (w.tags['oneway'] == 'yes'):
                # ONLY ONE DIRECTION
                G.add_path(w.nds, id=w.id, highway=w.tags['highway'])
            else:
                # BOTH DIRECTION
                G.add_path(w.nds, id=w.id, highway=w.tags['highway'])
                G.add_path(w.nds[::-1], id=w.id, highway=w.tags['highway'])
        else:
            # BOTH DIRECTION
            G.add_path(w.nds, id=w.id, highway=w.tags['highway'])
            G.add_path(w.nds[::-1], id=w.id, highway=w.tags['highway'])

    ## Complete the used nodes' information    
    for n_id in G.nodes.keys():
        n = osm.nodes[n_id]
        G.node[n_id]['lat'] = n.lat
        G.node[n_id]['lon'] = n.lon
        G.node[n_id]['id'] = n.id

    ## Estimate the length of each way
    for u,v in G.edges():
        distance = haversine(G.node[u]['lon'], G.node[u]['lat'], G.node[v]['lon'], G.node[v]['lat'], unit_m = True) # Give a realistic distance estimation (neither EPSG nor projection nor reference system are specified)

        G.add_weighted_edges_from([( u, v, distance)], weight='length')

    return G


class Node:
    def __init__(self, id, lon, lat):
        self.id = id
        self.lon = lon
        self.lat = lat
        self.tags = {}

    def __str__(self):
        return "Node (id : %s) lon : %s, lat : %s "%(self.id, self.lon, self.lat)


class Way:
    def __init__(self, id, osm):
        self.osm = osm
        self.id = id
        self.nds = []
        self.tags = {}

    def split(self, dividers):
        # slice the node-array using this nifty recursive function
        def slice_array(ar, dividers):
            for i in range(1,len(ar)-1):
                if dividers[ar[i]]>1:
                    left = ar[:i+1]
                    right = ar[i:]

                    rightsliced = slice_array(right, dividers)

                    return [left]+rightsliced
            return [ar]

        slices = slice_array(self.nds, dividers)

        # create a way object for each node-array slice
        ret = []
        i=0
        for slice in slices:
            littleway = copy.copy( self )
            littleway.id += "-%d"%i
            littleway.nds = slice
            ret.append( littleway )
            i += 1

        return ret

class OSM:
    def __init__(self, filename_or_stream):
        """ File can be either a filename or stream/file object."""
        nodes = {}
        ways = {}

        superself = self

        class OSMHandler(xml.sax.ContentHandler):
            @classmethod
            def setDocumentLocator(self,loc):
                pass

            @classmethod
            def startDocument(self):
                pass

            @classmethod
            def endDocument(self):
                pass

            @classmethod
            def startElement(self, name, attrs):
                if name=='node':
                    self.currElem = Node(attrs['id'], float(attrs['lon']), float(attrs['lat']))
                elif name=='way':
                    self.currElem = Way(attrs['id'], superself)
                elif name=='tag':
                    self.currElem.tags[attrs['k']] = attrs['v']
                elif name=='nd':
                    self.currElem.nds.append( attrs['ref'] )

            @classmethod
            def endElement(self,name):
                if name=='node':
                    nodes[self.currElem.id] = self.currElem
                elif name=='way':
                    ways[self.currElem.id] = self.currElem

            @classmethod
            def characters(self, chars):
                pass

        xml.sax.parse(filename_or_stream, OSMHandler)

        self.nodes = nodes
        self.ways = ways

        #count times each node is used
        node_histogram = dict.fromkeys( self.nodes.keys(), 0 )
        for way in self.ways.values():
            if len(way.nds) < 2:       #if a way has only one node, delete it out of the osm collection
                del self.ways[way.id]
            else:
                for node in way.nds:
                    node_histogram[node] += 1

        #use that histogram to split all ways, replacing the member set of ways
        new_ways = {}
        for id, way in self.ways.items():
            split_ways = way.split(node_histogram)
            for split_way in split_ways:
                new_ways[split_way.id] = split_way
        self.ways = new_ways
        
def fetch_roads_OSM(osm_path, acceptedRoads=["motorway","motorway_link","trunk","trunk_link","primary","primary_link","secondary","secondary_link","tertiary","tertiary_link"]):
    driver=ogr.GetDriverByName('OSM')
    data = driver.Open(osm_path)
        
    sql_lyr = data.ExecuteSQL("SELECT osm_id,highway FROM lines WHERE highway IS NOT NULL")
    
    roads=[]                          
    for feature in sql_lyr:
        if feature.GetField('highway') is not None:
            osm_id = feature.GetField('osm_id')
            shapely_geo = loads(feature.geometry().ExportToWkt()) 
            if shapely_geo is None:
                continue
            highway=feature.GetField('highway')
            if acceptedRoads != []:
                if highway in acceptedRoads:
                    roads.append([osm_id,highway,shapely_geo])
            else:
                roads.append([osm_id,highway,shapely_geo])
    if len(roads) > 0:
        road_gdf = gpd.GeoDataFrame(roads,columns=['osm_id','infra_type','geometry'],crs={'init': 'epsg:4326'})
        return road_gdf
    else:
        print('No roads in {}'.format(country))
        
def line_length(line, ellipsoid='WGS-84'):
    '''Length of a line in meters, given in geographic coordinates

    Adapted from https://gis.stackexchange.com/questions/4022/looking-for-a-pythonic-way-to-calculate-the-length-of-a-wkt-linestring#answer-115285

    Arguments:
        line {Shapely LineString} -- a shapely LineString object with WGS-84 coordinates
        ellipsoid {String} -- string name of an ellipsoid that `geopy` understands (see
            http://geopy.readthedocs.io/en/latest/#module-geopy.distance)

    Returns:
        Length of line in meters
    '''
    
    if line.geometryType() == 'MultiLineString':
        return sum(line_length(segment) for segment in line)

    return sum(
                vincenty(tuple(reversed(a)), tuple(reversed(b)), ellipsoid=ellipsoid).kilometers
                for a, b in pairwise(line.coords)
    )


import os, sys, logging, warnings, time

import osmnx
import networkx as nx
import pandas as pd
import geopandas as gpd
import numpy as np

from shapely.geometry import Point

#import GOSTnets as gn

def calculateOD_gdf(G, origins, destinations, fail_value=-1, weight="time", calculate_snap=False):
    ''' Calculate Origin destination matrix from GeoDataframes
    
    Args:
        G (networkx graph): describes the road network. Often extracted using OSMNX
        origins (geopandas dataframe): source locations for calculating access
        destinations (geopandas dataframe): destination locations for calculating access
        calculate_snap (boolean, optioinal): variable to add snapping distance to travel time, default is false
    Returns:
        numpy array: 2d OD matrix with columns as index of origins and rows as index of destinations
    '''
    #Get a list of originNodes and destinationNodes
    wgs84 = {'init':'epsg:4326'}
    if origins.crs != wgs84:
        origins = origins.to_crs(wgs84)
    if destinations.crs != wgs84:
        destinations = destinations.to_crs(wgs84)
    origins = gn.pandana_snap(G, origins)
    destinations = gn.pandana_snap(G, destinations)
    oNodes = origins['NN'].unique()
    dNodes = destinations['NN'].unique()
    od = gn.calculate_OD(G, oNodes, dNodes, fail_value)
    origins['OD_O'] = origins['NN'].apply(lambda x: np.where(oNodes==x)[0][0])
    destinations['OD_D'] = destinations['NN'].apply(lambda x: np.where(dNodes==x)[0][0])
    outputMatrix = od[origins['OD_O'].values,:][:,destinations['OD_D'].values]
    if calculate_snap:
        originsUTM = gn.pandana_snap(G, origins, target_crs='epsg:3857')
        destinationsUTM = gn.pandana_snap(G, destinations, target_crs='epsg:3857')
        originsUTM['tTime_sec'] =      originsUTM['NN_dist']      / 1000 / 5 * 60 * 60 # Convert snap distance to walking time in seconds
        destinationsUTM['tTime_sec'] = destinationsUTM['NN_dist'] / 1000 / 5 * 60 * 60 # Convert snap distance to walking time in seconds
        for idx, row in originsUTM.iterrows():
            outputMatrix_adj[idx,:] = outputMatrix_adj[idx,:] + row['tTime_sec']
        outputMatrix = outputMatrix_adj
    return(outputMatrix)    
    
def calculateOD_csv(G, originCSV, destinationCSV='', oLat="Lat", oLon="Lon", dLat="Lat", dLon="Lon", crs={'init':'epsg:4326'}, fail_value=-1, weight='time', calculate_snap=False):
    """
    Calculate OD matrix from csv files of points

    :param G: describes the road network. Often extracted using OSMNX
    :param string origins: path to csv file with locations for calculating access
    :param string destinations: path to csv with destination locations for calculating access
    :param string oLat:
    :param string oLon:
    :param string dLat:
    :param string dLon:
    :param dict crs: crs of input origins and destinations, defaults to {'init':'epsg:4326'}
    :param int fail-value: value to put in OD matrix if no route found, defaults to -1
    :param string weight: variable in G used to define edge impedance, defaults to 'time'
    :param bool calculate_snap: variable to add snapping distance to travel time, default is false
    :returns: numpy array: 2d OD matrix with columns as index of origins and rows as index of destinations
    """

    originPts = pd.read_csv(originCSV)
    pts = [Point(x) for x in zip(originPts[oLon],originPts[oLat])]
    originGDF = gpd.GeoDataFrame(originPts, geometry=pts, crs=crs)

    if destinationCSV == '':
        destinationGDF = originGDF.copy()
    else:
        originPts = pd.read_csv(destinationCSV)
        pts = [Point(x) for x in zip(originPts[dLon],originPts[dLat])]
        destinationGDF = gpd.GeoDataFrame(originPts, geometry=pts, crs=crs)
    OD = calculateOD_gdf(G, originGDF, destinationGDF, fail_value, weight, calculate_snap = calculate_snap)

    return(OD)
   
def calculate_gravity(od, oWeight=[], dWeight=[], decayVals=[0.01,
                                                        0.005,
                                                        0.001,
                                                        0.0007701635,   # Market access halves every 15 mins
                                                        0.0003850818,   # Market access halves every 30 mins
                                                        0.0001925409,   # Market access halves every 60 mins
                                                        0.0000962704,   # Market access halves every 120 mins
                                                        0.0000385082,   # Market access halves every 300 mins
                                                        0.00001]):
    ''' Calculate the gravity weight between origins and destinations.
    
    Args:
        od (ndarray): matrix of travel time    
        oWeight/dWeight (array, optional) - array of weights for calculating weight; reverts to 1 for if not defined   
        decayVals (array, optional): decayVals to calculate for gravity. Each value will be returned as a column of results
    Returns:
        geopandas: columns of decayvals
    '''
    if len(oWeight) != od.shape[0]:
        oWeight = [1] * od.shape[0]
    if len(dWeight) != od.shape[1]:
        dWeight = [1] * od.shape[1]
    allRes = []
    for dist_decay in decayVals:
        outOD = od * 0
        decayFunction = lambda x: np.exp(-1 * dist_decay * x)
        for row in range(0, od.shape[0]):
            curRow = od[row,:]
            decayedRow = decayFunction(curRow)
            weightedRow = decayedRow * oWeight[row] * dWeight
            outOD[row,:] = weightedRow
        summedVals = np.sum(outOD, axis=1)
        allRes.append(summedVals)
    res = pd.DataFrame(allRes).transpose()
    res.columns = columns=['d_%s' % d for d in decayVals]
    return(res)
    
    



from setuptools import setup, find_packages

setup(
    name="GOSTnets",
    version="1.0.0",
    packages=find_packages(),

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires=[#'docutils>=0.3',
                      'geopandas>=0.4.0',
                      'networkx',
                      'numpy',
                      'osmnx',
                      'pandas>=0.23.5',
                      'pyproj',
                      'scipy',
                      'shapely'
                      ],

    package_data={
        # If any package contains *.txt or *.rst files, include them:
        # And include any *.msg files found in the 'hello' package, too:
    },

    # metadata to display on PyPI
    author="Benjamin P. Stewart",
    author_email="ben.gis.stewart@gmail.com",
    description="Networkx wrapper to simplify network analysis using geospatial data",
    license="PSF",
    keywords="networkx networks OSM",
    url="https://github.com/worldbank/GOSTnets",   # project home page, if any
    project_urls={
        "Bug Tracker": "https://github.com/worldbank/GOSTnets/issues",
        "Documentation": "https://github.com/worldbank/GOSTnets",
        "Source Code": "https://github.com/worldbank/GOSTnets",
    }

    # could also include long_description, download_url, classifiers, etc.
)


import os, sys
#import rasterio

import pandas as pd
import geopandas as gpd
import numpy as np

from scipy import interpolate
#from rasterio import features
#from rasterio.mask import mask
#from rasterio.features import rasterize
#from rasterio.warp import reproject, Resampling

def rasterize_od_results(inD, outFile, field):
    ''' Convert gridded point data frame to raster of commensurate size and resolution
    
    INPUT
    inD [ geopandas data frame ] - OD matrix as point data frame
    outFile [ string ] - path to save output raster
    field [ string ] - field to rasterize
    
    RETURNS
    None
    '''
    #create grid from input shapefile
    # get xs, ys, and values from origin points
    xs = np.array(inD.geometry.apply(lambda p: p.x))
    ys = np.array(inD.geometry.apply(lambda p: p.y))
    vals = np.array(inD[field])
    # creates a full grid for the entire bounding box (all pairs of xs and ys)
    unique_xs = np.unique(xs)
    unique_ys = np.unique(ys)
    xx, yy = np.meshgrid(unique_xs, unique_ys)
    # this creates a new set of values to fill the grid
    grid_array = interpolate.griddata((xs,ys), vals, (xx, yy))
    x_pixels = grid_array.shape[1]
    y_pixels = grid_array.shape[0]
    # get the right transformation for raster file
    xRes = (xx.max() - xx.min()) / len(unique_xs)
    yRes = (yy.max() - yy.min()) / len(unique_ys)
    # get the right transformation for raster file
    trans = rasterio.transform.from_bounds(xx.min() - (xRes/2), yy.min() - (yRes/2), 
                                           xx.max() - (xRes/2), yy.max() - (yRes/2),
                                           x_pixels - 1, y_pixels - 1)
    new_dataset = rasterio.open(
        outFile, 'w', driver = 'GTiff',
        height = y_pixels, width = x_pixels,
        count=1, dtype=str(grid_array.dtype),
        crs=inD.crs,
        transform=trans
    )
    
    shapes = ((row.geometry,row[field]) for idx, row in inD.iterrows())
    burned = features.rasterize(shapes=shapes, fill=0, out_shape=grid_array.shape, transform=new_dataset.transform)
    burned = burned.astype(grid_array.dtype)    
    new_dataset.write_band(1, burned)
    
    new_dataset.close()